import numpy as np
from bitarray import bitarray
from functools import partial
from typing import Callable, List, Union
from itertools import groupby
import re

from .grid import Grid
from .similarity import tanimoto_similarity

from scipy.constants import electron_volt


class DOSFingerprint():
    """
    A fingerprint of the electronic density-of-states (DOS), 
    obtained by integrating the DOS over discrete intervals
    and binning of the resulting histogram.
    
    **Keyword arguments:**

    similarity_function: `Callable`
        A function that allows to calculate the similarity between two DOS fingerprints    

        default: `nomad_dos_fingerprints.similarity.tanimoto_similarity`

    Additional keyword arguments are passed to the similarity function.
    """

    def __init__(self, 
                 similarity_function: Callable = tanimoto_similarity, **kwargs):
        self.bins = ''
        self.indices = []
        self.filling_factor = 0
        self.grid_id = None
        self.set_similarity_function(similarity_function, **kwargs)
        self._n_state_bins = None
        self._bitarray = None

    @property
    def n_state_bins(self):
        if self._n_state_bins is None:
            assert self.grid_id is not None, "Grid id is not set. Will be set when calculating the fingerprint with a grid."
            self._n_state_bins = Grid.resolve_grid_id(self.grid_id)["n_pix"]
        return self._n_state_bins

    def calculate(self, 
                  dos_energies: np.ndarray, 
                  dos_values: np.ndarray,
                  grid: Grid | None = None, 
                  grid_id: str = None, 
                  convert_data: Union[None, str, Callable] = None, 
                  normalization_factor: float = 1.0) -> object:
        """
        Calculate the fingerprint data from normalized DOS energies and values.

        **Arguments:**

        dos_energies: `np.ndarray`
            Energies of the DOS spectrum

            By default, units are [eV].

            This behavior can be changed by changing `convert_data` (see below).

        dos_values: `np.ndarray`
            Values of the DOS spectrum

            By default, units are [states/cell/eV].

            This behavior can be changed by changing `convert_data` (see below).

        **Keyword arguments:**

        grid: `Grid` object
            Grid for calculating the fingerprint.  
            For details, see documentation there.

            If both `grid` and `grid_id` are given and they don't represent the same grid, a `ValueError` is raised.

            default: `None`

        grid_id: `str`
            ID for describing the `Grid` object that is used to calculate the fingerprint data.
            For details, see documentation there.

            If both `grid` and `grid_id` are given and they don't represent the same grid, a `ValueError` is raised.

            default: `None`: Use `Grid` default values.

        convert_data: `Union[str, None, Callable]`
            Convert `dos_energies` and `dos_values` to different units.

            The string `'enc'` transforms [Joule] to [eV] and sums all spin channels of DOS values. 
            Additionally, assumes that `dos_energies` is an `Interable` of `float`s 
            and `dos_values` is an `Interable` of `Interables` of lenght `len(dos_energies)` of `float`s.

            If set to `None`, no conversion will be performed.

            If a callable is given, it will be called as:
            ``energy, dos = convert_data(dos_energies, dos_values)``.

            default: `None`

        normalization_factor: `float`
            Factor used for unit conversion, e.g. if the DOS is given or required per atom.
            Used only if `convert_data` is "Enc".

            default: `1.0`

        **Returns:**

        self: `DOSFingerprint`
            Fingerprint after calculation
        """
        if isinstance(convert_data, str) and convert_data.lower() == 'enc':
            energy, dos = self._convert_dos(dos_energies, dos_values, normalization_factor = normalization_factor)
        elif isinstance(convert_data, Callable):
            energy, dos = convert_data(dos_energies, dos_values)
        elif convert_data is None:
            energy, dos = dos_energies, dos_values
        else:
            raise ValueError('Key-word argument `convert_data` must be either the string "enc", a callable, or `None`.')
        if grid is None:
            grid = Grid.create(grid_id = grid_id)
        else:
            if grid_id is not None and grid.get_grid_id() != grid_id:
                raise ValueError('Both grid and grid_id are given and do not correspond to the same grid.')
        energy = np.array(energy) - grid.e_ref
        raw_energies, raw_dos = self._integrate_to_bins(energy, dos, grid.delta_e_min)
        self.grid_id = grid_id if grid_id is not None else grid.get_grid_id()
        bin_fp = self._calculate_bit_fingerprint(raw_energies, raw_dos, grid)
        self._bitarray = bitarray(bin_fp)
        self.bins = self._compress_binary_fingerprint_string(bin_fp)
        self._n_state_bins = grid.n_pix
        return self

    def to_dict(self) -> dict:
        """
        Convert data to dictionary.
        """
        data_dict = {}
        for key in ['bins', 'indices', 'grid_id', 'filling_factor', 'overflow', 'undersampling']:
            data_dict[key] = getattr(self, key)
        return data_dict

    @classmethod
    def from_dict(cls, fp_dict: dict) -> object:
        """
        Create fingerprint object from dictionary.
        """
        self = cls()
        self.bins = fp_dict['bins']
        self.indices = fp_dict['indices']
        self.grid_id = fp_dict['grid_id']
        self.filling_factor = fp_dict['filling_factor']
        if 'overflow' in fp_dict.keys():
            self.overflow = fp_dict['overflow']
        if 'undersampling' in fp_dict.keys():
            self.undersampling = fp_dict['undersampling']
        self._n_state_bins = Grid.resolve_grid_id(self.grid_id)["n_pix"]
        return self

    def set_similarity_function(self, similarity_function: Callable, **kwargs):
        """
        Set the similarity function of the fingerprint.

        **Arguments:**

        similarity_function: `Callable`
            Function for calculating the similarity between `DOSFingerprint` objects.

        Keyword arguments are passed to the similarity function.
        """
        self.similarity_function = partial(similarity_function, **kwargs)

    def get_similarity(self, fingerprint: object) -> float:
        """
        Get similarity value between self and another fingerprint object.

        **Arguments:**

        fingerprint: `DOSFingerprint`
            Other fingerprint to calculate similarity to.

        **Returns:**

        similarity: `float`
            Similarity between both fingerprints.

        **Raises:**

        `TypeError`: `fingerprint` is not a `DOSFingerprint`
        """
        if not isinstance(fingerprint, DOSFingerprint):
            raise TypeError("Other fingerprint must be a `DOSFingerprint` object.")
        return self.similarity_function(self, fingerprint)

    def get_similarities(self, list_of_fingerprints: List[object]) -> np.ndarray:
        """
        Get similarities of self to a list of other fingerprints.

        **Arguments:**

        list_of_fingerprints: `List[DOSFingerprint]`
            Other fingerprints to calculate similarity to.

        **Returns:**

        similarities: `np.ndarray`
            Similarities of self to other fingerprints.
        """
        return np.array([self.similarity_function(self, fp) for fp in list_of_fingerprints])

    def get_bitarray(self) -> object:
        """
        Get `bitarray.bitarray` representing the fingerprint data.

        **Returns:**

        bits: `bitarray`
            Binary fingerprint data 
        """
        if self.bins == '':
            raise AttributeError("Fingerprint is not calculated! Use `calculate()`.")
        if self._bitarray is not None:
            return self._bitarray
        bit_string = self._expand_fingerprint_string(self.bins)
        bits = bitarray(bit_string)
        self._bitarray = bits
        return bits

    def __eq__(self, other):
        if not isinstance(other, DOSFingerprint):
            return False
        for attr in ['bins', 'indices', 'stepsize', 'grid_id', 'filling_factor']:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    def _integral_house(self, a, b, interval):
        # integrates a "house" like figure, consisting of a rectangle and a 90 degree triangle, with heights a and b, and width=interval
        return (a + b) * interval / 2.

    def _interpolate_dos(self, dos_values, energy_values, requested_energy):
        """
        Returns a linearly interpolated value between two DOS points.
        """
        if len(dos_values) != 2 or len(energy_values) != 2:
            raise ValueError("Error in _interpolate_dos: Wrong number of arguments for calculation of gradient.")
        if energy_values[0] == energy_values[1]:
            return dos_values[0]
        gradient = (dos_values[1] - dos_values[0]) / ((energy_values[1] - energy_values[0]) * 1.)
        difference = requested_energy - energy_values[0]
        interpolated = dos_values[0] + gradient * difference
        return interpolated
    
    def _find_energy_cutoff_indices(self, energy, estart, estop):
        """
        This find the correct indices for the integration, that is, all values of the energy in the interval [estart-1,estop+1] will be in the range [emin_idx,emax_idx]
        WARNING: Assumes that estart<energy[0]!
        """
        emin_idx = None
        emax_idx = None
        index = 0
        while emin_idx is None or emax_idx is None:
            if energy[index] > estart and emin_idx is None:
                emin_idx = index - 1
            if energy[-(index + 1)] < estop and emax_idx is None:
                emax_idx = -index + 1
                if emax_idx > 0:
                    emax_idx = 0 #HOTFIX
            index += 1
        if emax_idx == 0:
            emax_idx = None
        return emin_idx, emax_idx

    def _integrate_to_bins(self, energy, dos, stepsize=0.05):
        """
        Performs stepwise numerical integration of ``ys`` over the range of ``xs``.
        """
        if len(energy) < 2 or len(dos) < 2:
            raise ValueError(f'Invalid input. Please provide arrays with len > 2. len(x) : {len(energy)} len(y) : {len(dos)}')
        
        # define the limits that fit with the predefined stepsize 
        xstart = round(np.ceil(energy[0] / (stepsize * 1.)) * stepsize, 12)  
        xstop = round(np.floor(energy[-1] / (stepsize * 1.)) * stepsize, 12)

        # Find the indices in original arrays closest to new limits defined by stepsize
        idx_min, idx_max = self._find_energy_cutoff_indices(energy, xstart, xstop)

        energy = energy[idx_min:idx_max]
        dos = dos[idx_min:idx_max]
        current_energy = xstart
        current_dos = self._interpolate_dos([dos[0], dos[1]], [energy[0], energy[1]], xstart)
        dos_binned = []
        energy_binned = []
        index = 1  # starting from the second value, because energy[0]<xstart. Thus energy[index] is made to be larger than current_energy.
        while current_energy < xstop:
            energy_binned.append(current_energy)
            next_energy = round(current_energy + stepsize, 9)
            integral = 0
            while energy[index] < next_energy and next_energy < xstop:
                integral += self._integral_house(current_dos, dos[index], energy[index] - current_energy)
                current_dos = dos[index]
                current_energy = energy[index]
                index += 1
            next_dos = self._interpolate_dos([dos[index - 1], dos[index]], [energy[index - 1], energy[index]],
                                             next_energy)
            integral += self._integral_house(current_dos, next_dos, next_energy - current_energy)
            dos_binned.append(integral)
            current_energy = next_energy
            current_dos = next_dos
        return energy_binned, dos_binned

    def _convert_dos(self, energy, dos, normalization_factor = 1):
        """
        Convert units of DOS from energy: Joule; dos: states/volume/Joule to eV and sum spin channels if they are present.
        """
        energy = np.array([value / electron_volt for value in energy])
        dos_channels = [np.array(values) for values in dos]
        dos = sum(dos_channels) * electron_volt * normalization_factor
        return energy, dos

    def _binary_bin(self, dos_value: float, grid_bins: np.ndarray):
        bin_dos = ''
        for grid_bin in grid_bins:
            if grid_bin <= dos_value:
                bin_dos += '1'
            else:
                bin_dos += '0'
        return bin_dos

    def _calc_grid_indices(self, energy_bins: np.ndarray, grid: Grid):
        """
        Calculate indices of the DOS grid that describe the energy range where the DOS is defined.
        Args:
            energy_bins: np.ndarray: energy bins cut to be no larger than the Grid range
            grid: Grid: description of the DOS discretization
        Return:
            self.indices: list: Grid indices that indicate which parts of the Grid are no larger than the energy bins
        """
        grid_start, grid_end = grid.get_grid_indices_for_energy_range(energy_bins)
        self.indices = [grid_start, grid_end]
        return self.indices
        
    def _adapt_energy_bin_sizes(self, energy_bins_raw: np.ndarray, states_raw: np.ndarray, grid: Grid):
        """
        Adapt energy bin sizes to the specified values in the grid.
        Args:
            energy_bins_raw: numpy.ndarray: locations of energy bins
            states_raw: numpy.ndarray: states in the bins declared in `energy_bins`
            grid: Grid: description of the DOS discretization
        Returns:
            adapted_bins: list: new energy discretization steps
            adapted_states: list: state bins with adapted discretization steps
        """
        grid_array = grid.grid()
        # cut the energy and states to grid size
        energy_bins, states = np.transpose([(e,d) for e,d in zip(energy_bins_raw, states_raw) if (e >= grid_array[0][0] and e <= grid_array[-1][0])])
        # find grid start and end points
        grid_start, grid_end = self._calc_grid_indices(energy_bins, grid)
        # if the last energy bin is less than or equal to the last grid point, the pre-to-last grid index was found instead of the correct one
        # the following lines fix this, but it is not very elegant
        if grid_end <= len(grid_array) - 1:
            if len(energy_bins) == len(energy_bins_raw) and energy_bins_raw[-1] > grid_array[grid_end][0]:
                grid_end += 1
                self.indices[1] += 1
        # sum dos bins to adapt to inhomogeneous energy grid
        adapted_bins = []
        adapted_states = []
        for index in range(grid_start, grid_end):
            eps_i = grid_array[index][0]
            eps_iplusdelta = grid_array[index+1][0]
            adapted_bins.append(eps_i)
            adapted_states.append(sum(np.array([s for e, s in zip(energy_bins, states) if e >= eps_i and e < eps_iplusdelta])))
        return adapted_bins, adapted_states

    def _calc_bit_vector(self, adapted_states: np.ndarray, grid: Grid):
        """
        Calculate bit representation of grid-adapted states per energy bin.
        Args:
            adapted_states: numpy.ndarray: states in the bins declared in `energy_bins`
            grid: Grid: description of the DOS discretization
        Sets:
            overflow: float: sum of states that can not be described by the current grid
            undersampling: float: fraction of states that are in partially filled grid cells (and thus ignored)
        Returns: 
            discrete_states: str: binary representation of binned DOS spectrum
        """
        bin_fp = ''
        grid_array = grid.grid()
        grid_start, grid_end = self.indices
        overflow = 0
        undersampling = 0
        for states, grid_segment in zip(adapted_states, grid_array[grid_start:grid_end + 1]):
            bin_fp += self._binary_bin(states, grid_segment[1])
            overflow += self._get_segment_overflow(states, grid_segment)
            undersampling += self._get_segment_undersampling(states, grid_segment)
        self.overflow = overflow
        self.undersampling = undersampling / sum(adapted_states)
        return bin_fp

    def _calculate_byte_representation(self, bin_fp: str):
        byte_fp = bitarray(bin_fp).tobytes().hex()
        return byte_fp

    def _compress_group(self, count: int, char: str):
        representative_char = 't' if char == '1' else 'f'
        return str(count) + representative_char

    def _calculate_bit_fingerprint(self, binned_energies: np.ndarray, binned_states: np.ndarray, grid: Grid):
        """
        Calculate bit representation of DOS fingerprint.
        """
        _, adapted_states = self._adapt_energy_bin_sizes(binned_energies, binned_states, grid)
        bin_fp = self._calc_bit_vector(adapted_states, grid)
        self.filling_factor = bin_fp.count('1') / len(bin_fp)
        return bin_fp

    def _get_segment_overflow(self, states, grid_segment):
        """
        Calculate the number of states that lies beyond the region described by a grid segment.
        Args:
            states: float: states in the bins declared in `energy_bins`
            grid_segment: Grid: description of the DOS discretization
        Returns:
            segment_overflow: float: number of states that can be described by the grid segment
        """
        delta_states = states - grid_segment[1][-1] 
        segment_overflow = delta_states if delta_states >= 0 else 0
        return  segment_overflow
    
    def _get_segment_undersampling(self, states, grid_segment):
        """
        Calculate the amount of states that are in partially filled grid cells.
        Args:
            states: float: states in the bins declared in `energy_bins`
            grid_segment: Grid: description of the DOS discretization
        Returns:
            segment_undersampling: float: sum of states that are in partially filled grid cells

        """
        delta_states = 0
        for idx in range(len(grid_segment[1]) - 1):
            if grid_segment[1][idx] < states < grid_segment[1][idx+1]:
                delta_states = states - grid_segment[1][idx]
                break
        return delta_states

    def _compress_binary_fingerprint_string(self, fingerprint_string: str):
        compressed_string = ''
        for char, count in groupby(fingerprint_string):
            compressed_string += self._compress_group(len(list(count)), char)
        return compressed_string

    def _expand_fingerprint_string(self, compressed_fingerprint_string):
        decompressed_string = ''
        for bits in re.findall(r"(\d+\w)", compressed_fingerprint_string):
            value = '1' if bits[-1] == 't' else '0'
            decompressed_string += value * int(bits[:-1])
        return decompressed_string


from madas import Fingerprint, Material

import pytest

@pytest.fixture()
def material():
    return Material("a")

def test_from_material(material):

    prop = {"test" : {"path": 1}}
    material.set_data(prop)

    fp = Fingerprint("PROP", property_path = "test/path").from_material(material)

    assert fp.get_similarity(fp) == 1, "Same fingerprint is not identical"

    prop = {"test" : {"path": 2}}
    material.set_data(prop)

    fp2 = Fingerprint("PROP", property_path = "test/path").from_material(material)

    assert fp.get_similarity(fp2) == 0.5, "Wrong similarity between prop fingerprints"

def test_similarity(material):

    prop = {"test" : {"path": 1}}
    material.set_data(prop)

    fp = Fingerprint("PROP", property_path = "test/path", pass_on_exceptions=False).from_material(material)

    prop = {"test1" : {"path": 2}}
    material.set_data(prop)

    fp2 = Fingerprint("PROP", property_path = "test1/path", pass_on_exceptions=False).from_material(material)

    print(fp.data["property_path"])
    print(fp2.data["property_path"])

    with pytest.raises(ValueError):
        fp.get_similarity(fp2)


from madas.apis.api_core import APIClass
from madas.backend.backend_core import Backend
from madas.fingerprint import Fingerprint
import pytest
from madas import Material, MaterialsDatabase
from ase.build import bulk

from madas.similarity import SimilarityMatrix

class MockAPI(APIClass):

    def __init__(self, test_material) -> None:
        self._test_material = test_material
        self._called_get_calculation = 0
        self._called_get_property = 0
        self._called_get_calculations_by_search = 0

    def get_calculations_by_search(self, some_string, **kwargs):
        """
        Test doc string
        """
        if some_string == "duplicates":
            self._called_get_calculations_by_search += 1
            return [self.test_material, self.test_material]
        self._called_get_calculations_by_search += 1
        return [self.test_material]

    @property
    def test_material(self):
        return self._test_material

    def get_calculation(self, return_id = 1, *args, **kwargs):
        """
        Test doc string
        """
        self._called_get_calculation += 1
        new_material = self.test_material
        if return_id == 2:
            new_material.mid = "c:d"
        return new_material

    def get_property(self, *args, **kwargs):
        """
        Test doc string
        """
        self._called_get_property += 1
        return 1

    def set_logger(self, logger):
        self.log = logger

    def _gen_mid(self, return_id = 1):
        if return_id == 1:
            return "a:b"
        else:
            return "c:d"

    def resolve_mid(self, *args):
        return {"ignore" : "this"}

class MockFingerprint():

    def __init__(self, *args, name = None, **kwargs) -> None:
        self.calculated = False
        self.set_fp_type("Mock")
        self.set_name(name)
        self._mid = "a:b"
        self._data = {"test" : "data"}

    def calculate(self, *args, **kwargs) -> object:
        self.calculated = True
        return self
    
    def from_material(self, *args, **kwargs) -> object:
        self.calculate( *args, **kwargs)
        return self

    def from_data(self, *args, **kwargs) -> object:
        return MockFingerprint().calculate()

class MockBackend(Backend):

    def __init__(self, filename="materials_database.db", filepath="data", make_dirs=True, key_name="mid", log=None):
        super().__init__(filename, filepath, make_dirs, key_name, log)
        self._added_single = 0
        self._added_many = 0
        self._update_buffer = []

    def add_single(self, *args, **kwargs):
        self._added_single += 1

    def add_many(self, entries, **kwargs):
        if len(set(entries)) < len(entries):
            raise ValueError("Tried to write same entry to db twice!") 
        self._added_many += 1

    def get_by_id(self, db_id):
        return Material("a:b", atoms=bulk('Cu', 'fcc', a=3.6), data = {"test" : "data", "dos" : {"a" : "a", "b" : "b"}}, properties = {"test_prop" : "property"})

    def get_single(self, mid=None, **kwargs):
        return self.get_by_id(None)

    def has_entry(self, entry_id):
        if self._added_single or self._added_many:
            return True
        return False

    def get_length(self):
        return self._added_single + self._added_many

    def update_many(self, mids, kwargs_list):
        self._update_buffer.append([mids, kwargs_list])

    def update_single(self, mid, **kwargs):
        self._update_buffer.append([mid, kwargs])

    def update_metadata(self, *args, **kwargs):
        self._metadata.update(**kwargs)

class MockSimilarityMatrix():

    def __init__(self, calculated = False, **kwargs) -> None:
        self._calculated = calculated

    def calculate(self, *args, **kwargs):
        return MockSimilarityMatrix(True)

@pytest.fixture()
def test_atoms():
    return bulk('Cu', 'fcc', a=3.6)

@pytest.fixture()
def test_query():
    return {
        "search_by":
            {
                "elements": ["Cu","Pt"], 
                "exclusive":True, 
                "restricted":True,
                "page":1,
                "per_page":10
            },
        "has_dos":True,
        "testquery" : "query"
    }

@pytest.fixture()
def test_material(test_atoms):
    return Material("a:b", atoms=test_atoms, data = {"test" : "data"})

@pytest.fixture()
def materials_database(tmpdir, test_material):
    db = MaterialsDatabase(api = MockAPI(test_material), backend=MockBackend(filepath=tmpdir), filepath=tmpdir, log_mode="stream")
    return db

def test_setup(materials_database):

    assert materials_database.log.name == "materials_database_log"
    assert materials_database.api_logger.name == "materials_database_api"

def test_API_missing_docstrings(tmpdir):

    class EmptyAPI(APIClass):

        def get_calculation(self, *args, **kwargs) -> Material:
            return Material("a")

        def get_calculations_by_search(self, *args, **kwargs):
            return [Material("a")]

        def get_property(self, **kwargs):
            return "a"

    MaterialsDatabase(filepath=tmpdir, api=EmptyAPI(), log_mode="stream")

def test_add_material(materials_database):
    
    materials_database.add_material()

    assert materials_database.api._called_get_calculation == 1, "Did not make correct API function call"
    assert materials_database.backend._added_single == 1, "Did not call Backend.add_single function"

    materials_database.add_material()

    assert materials_database.api._called_get_calculation == 1, "Tried to get materials existing material"
    assert materials_database.backend._added_single == 1, "Tried to store existing material"

def test_get_property(materials_database, test_material):
    
    materials_database.add_material()

    assert materials_database.get_property(test_material.mid, "test") == "data", "Unable to get property from Material().data"

    assert materials_database.get_property(test_material.mid, "test_prop") == "property", "Unable to get property from Material().properties"

    assert materials_database.get_property(test_material.mid, "dos/a") == "a", "Unable to get nested property from Material().data"

def test_iteration(materials_database):

    assert len(materials_database) == len(list(materials_database)), "Returned more materials than available in database"

def test_get_properties(materials_database, test_material):

    materials_database.add_material()

    properties, mids = materials_database.get_properties("dos/a", True)
    assert properties == ["a"], "Did not recieve correct list of properties"
    assert mids == [test_material.mid], "Did not recieve correct list of mids"

def test_get_property_dataframe(materials_database):

    materials_database.add_material()

    import pandas as pd

    ref_dataframe=pd.DataFrame({'dos/a': {'a:b': 'a'}, 'dos/b': {'a:b': 'b'}, 'test': {'a:b': 'data'}})

    assert ((materials_database.get_property_dataframe(["dos/a", "dos/b", "test"]) == ref_dataframe).all()).all(), "Did not create correct dataframe"

def test_getitem(materials_database):
    materials_database.add_material()
    assert materials_database["a"].mid == "a:b", "Did not recieve correct material by mid"
    assert materials_database[0].mid == "a:b", "Did not recieve correct material by integer"

def test_get_fingerprint(materials_database, monkeypatch):
    
    monkeypatch.setattr(Fingerprint, "__init__", MockFingerprint.__init__)
    monkeypatch.setattr(Fingerprint, "calculate", MockFingerprint.calculate)
    monkeypatch.setattr(Fingerprint, "from_data", MockFingerprint.from_data)

    fp = materials_database.get_fingerprint("Mock", "a")

    assert fp.fp_type == "Mock", "Calulated wrong fingerprint"
    assert fp.mid == "a:b", "Wrong mid for fingerprint"

def test_get_fingerprints(materials_database, monkeypatch):

    materials_database.add_material()

    monkeypatch.setattr(Fingerprint, "__init__", MockFingerprint.__init__)
    monkeypatch.setattr(Fingerprint, "calculate", MockFingerprint.calculate)
    monkeypatch.setattr(Fingerprint, "from_data", MockFingerprint.from_data)

    fps = materials_database.get_fingerprints("Mock", "a")

    assert len(fps) == 1, "Wrong length of fingerprint list"
    assert fps[0].fp_type == "Mock", "Calulated wrong fingerprint"
    assert fps[0].mid == "a:b", "Wrong mid for fingerprint"

def test_get_similarity_matrix(materials_database, monkeypatch):
    
    monkeypatch.setattr(SimilarityMatrix, "__init__", MockSimilarityMatrix.__init__)
    monkeypatch.setattr(SimilarityMatrix, "calculate", MockSimilarityMatrix.calculate)

    monkeypatch.setattr(Fingerprint, "__init__", MockFingerprint.__init__)
    monkeypatch.setattr(Fingerprint, "calculate", MockFingerprint.calculate)
    monkeypatch.setattr(Fingerprint, "from_data", MockFingerprint.from_data)

    materials_database.add_material()

    simat = materials_database.get_similarity_matrix(None)

    assert simat._calculated is True, "Similarity matrix was not calculated"

def test_add_fingerprint(materials_database, monkeypatch):

    monkeypatch.setattr(Fingerprint, "__init__", MockFingerprint.__init__)
    monkeypatch.setattr(Fingerprint, "calculate", MockFingerprint.calculate)
    monkeypatch.setattr(Fingerprint, "from_data", MockFingerprint.from_data)

    materials_database.add_material()

    materials_database.add_fingerprint("Mock", name = "Mock1")

    assert materials_database.backend._update_buffer == [[['a:b'], [{'Mock1': '{"test": "data"}'}]]], "Tried to update with wrong data"

    assert materials_database.get_metadata()["fingerprints"] == ["Mock1"], "Did not write correct metadata"

def test_add_fingerprint_by_type(materials_database):

    class TestFingerprint(Fingerprint):

        def calculate(self, *args, **kwargs):
            self.set_data("calculated", {"test" : 1})
            return self
        
        def from_material(self, material: Material, *args, **kwargs):
            self.set_mid(material)
            self.calculate(*args, **kwargs)
            return self


    def Test_similarity(fp1, fp2):
        return fp1.data["calculated"]["test"] + fp2.data["calculated"]["test"]

    materials_database.add_material()

    materials_database.add_fingerprint(TestFingerprint)

    fps = materials_database.get_fingerprints(TestFingerprint, similarity_function = Test_similarity)

    assert fps[0].mid == "a:b", "Did not load fingerprint mid"

    assert fps[0].get_similarities(fps) == [2], "Did not calculate similarities for deserialized fingerprint."

def test_add_fingerprints(materials_database, monkeypatch):

    monkeypatch.setattr(Fingerprint, "__init__", MockFingerprint.__init__)
    monkeypatch.setattr(Fingerprint, "calculate", MockFingerprint.calculate)
    monkeypatch.setattr(Fingerprint, "from_data", MockFingerprint.from_data)

    materials_database.add_material()

    materials_database.add_fingerprints(["Mock", "Mock"], ["m1", "m2"])

    assert materials_database.backend._update_buffer == [[['a:b'], [{'m1': '{"test": "data"}', 'm2' : '{"test": "data"}'}]]], "Tried to update with wrong data"

    assert materials_database.get_metadata()["fingerprints"] == ["m1", "m2"], "Did not write correct metadata"

def test_fill_database(materials_database, caplog):
    
    materials_database.fill_database({"a" : "a"})

    assert materials_database.backend._added_many == 1, "Did not add many to backend"
    assert materials_database.api._called_get_calculations_by_search == 1, "Did not call API"
    assert materials_database.get_metadata()["search_queries"] == ['bd3ab8c72d54553816132940bffd738c']

    materials_database.fill_database({"a" : "a"})

    assert materials_database.backend._added_many == 1, "Tried to add redundant materials"
    assert materials_database.api._called_get_calculations_by_search == 1, "Queried API twice"

    materials_database.fill_database({"b" : "b"})

    assert materials_database.backend._added_many == 2, "Tried to add redundant materials"
    assert materials_database.api._called_get_calculations_by_search == 2, "Queried API twice"

    assert "Material a:b already in database. Skipping." in str(caplog.text), "Did not avoid adding duplicates"

def test_fill_database_no_duplicates_from_api(materials_database):

    materials_database.fill_database("duplicates")

def test_get_random(materials_database):

    materials_database.add_material()

    assert materials_database.get_random(return_mid = True) == "a:b", "Did not get material from get_random" 

def test_update_entry(materials_database):

    materials_database.update_entry("a", b = "c")

    assert materials_database.backend._update_buffer == [["a", {"b" : "c"}]], "Wrong data passed to update function"

def test_update_entries(materials_database):

    materials_database.update_entries(["a"], [{"b" : "c"}])

    assert materials_database.backend._update_buffer == [[["a"], [{"b" : "c"}]]], "Wrong data passed to update function"

def test_add_property(materials_database):

    materials_database.add_property("a", "b")

    assert materials_database.api._called_get_property == 1, "Did not call API"
    assert materials_database.backend._update_buffer == [['a', {"data" : {"b" : 1}}]], "Did not update data correctly"

def test_update_metadata(tmpdir):

    db = MaterialsDatabase(filepath=tmpdir, log_mode="stream")

    db._update_metadata({"test":"this"})

    del db

    db = MaterialsDatabase(filepath=tmpdir, log_mode="stream")

    assert db.get_metadata() == {'database_name': 'unnamed', "test":"this"}, "Did not recover metadata"



import pytest
from madas.apis.NOMAD_web_API import get_atoms

import numpy as np

@pytest.fixture()
def mock_response():
    data = {
        "archive" : {
            "results" : {
                "properties" : {
                    "structures" : {
                        "structure_original" : {
                            "cartesian_site_positions" : [
                                [0,0,0],
                                [25e-12,25e-12,25e-12]
                            ],
                            "lattice_vectors" : [
                                [1e-10, 0, 0],
                                [0, 1e-10, 0],
                                [0, 0, 1e-10]
                            ],
                            "species_at_sites" : [
                                "C", "Si"
                            ]
                        }
                    }
                }
            },
            "mock_data" : "some_test_data",
            "run" : [
                {"system" : [{
                    "atoms" : {
                        "periodic" : [True, True, True]
                    }
                }]}
            ]
        }
    }
    return data

def test_get_atoms(mock_response):

    at = get_atoms(mock_response)

    assert np.allclose(at.positions, [[0,0,0], [.25,.25,.25]]), "Failed to parse atomic positions"
    assert np.allclose(at.cell, np.eye(3)), "Failed to parse atomic cell"
    assert (at.symbols == ["C", "Si"]).all(), "Failed to parse atomic symbols"




from madas.fingerprints.DUMMY_fingerprint import DUMMYFingerprint

def test_dummy_fingerprint():
    data = [1,2,3,4]
    fp = DUMMYFingerprint().calculate(data)
    fp.set_pass_on_exceptions(True)
    assert fp.get_similarity(fp) == 1, "Wrong similarity"
    assert fp.y == 30, "wrong target value"



from .DOS_fingerprint import DOSFingerprint, DOS_similarity  # noqa: F401
from .DUMMY_fingerprint import DUMMYFingerprint, DUMMY_similarity  # noqa: F401
from .PROP_fingerprint import PROPFingerprint, PROP_similarity  # noqa: F401
from .PTE_fingerprint import PTEFingerprint, PTE_similarity  # noqa: F401
from .SYM_fingerprint import SYMFingerprint, SYM_similarity  # noqa: F401


from ase.spacegroup import get_spacegroup
from ase import Atoms

from madas import Fingerprint
from madas.material import Material


def SYM_similarity(fingerprint1: Fingerprint, fingerprint2: Fingerprint) -> float:
    """
    Compute the similarity between two `SGFingerprint` objects.

    Similarily to the Tanimoto coefficient, the SG similarity is defined as the 
    intersection of symmetry operations of two materials, divided by their union. 
    """
    symops1 = set(fingerprint1.data["symops"])
    symops2 = set(fingerprint2.data["symops"])

    return len(symops1.intersection(symops2)) / len(symops1.union(symops2))


class SYMFingerprint(Fingerprint):
    """
    SYMmetry fingerprint, defined by the symmetry operations of a crystal lattice.

    These are determined using `ASE`s `spacegroup` module. 
    """


    def __init__(self, 
                 name = "SYM", 
                 similarity_function = SYM_similarity,
                 pass_on_exceptions = True) -> None:
        self.set_fp_type("SYM")
        self.set_name(name)
        self.set_similarity_function(similarity_function)
        self.set_pass_on_exceptions(pass_on_exceptions)

    def from_material(self, 
                      material: Material, 
                      symprec: float = 1e-1, 
                      unify_species: bool = False) -> object:
        """
        *Calculatate* fingerprint from a `Material` object.

        See `calculate` method for a description of the parameters.
        """
        self.set_mid(material)
        atoms = material.atoms
        return self.calculate(atoms, symprec=symprec, unify_species=unify_species)

    def calculate(self, 
                  atoms: Atoms, 
                  symprec: float = 1e-1, 
                  unify_species: bool = False) -> object:
        """
        Calculate the SG fingerprint values. Uses `ASE`s `spacegroup` module to compute the space group.

        **Arguments:**

        atoms: `ase.Atoms`
            ASE Atoms object that contains (periodic) structure to determine the symmetry operations.

        **Keyword arguments:**

        symprec: `float`
            Precision, i.e. tolerance, for atom positions

            default: `0.1`

        unify_species: `bool`
            Set all species to the same value before computing symmetry operations.
            
            default: `False`
        """
        atoms = atoms.copy()
        if unify_species:
            atoms.set_atomic_numbers([1 for _ in atoms.get_atomic_numbers()])
        sg_ = get_spacegroup(atoms, symprec=symprec)
        self.set_data("SGN", sg_.no)
        self.set_data("symops", list(map(self._symop_to_string, sg_.get_symop())))
        return self

    def _symop_to_string(symop):
        rot = symop[0].flatten().tolist()
        rot.extend(symop[1])
        return ":".join(map(str, rot))



import logging
import os
from typing import List

from madas import Material

class Backend():
    """
    A database backend wrapper to unify file-based storage of materials data. 

    **Keyword arguments:**

    filename: `str`
        Name of the data file

        default: `"materials_database.db"`

    filepath: `str`
        Path (location) of the file.

        default: `"data"`

    make_dirs: `bool`
        Create directory paths if it does not exist.

        default: `True`

    key_name: `str`
        Name of unique keys that are used to find (individual) entries in the database.

        default: `"mid"`
    
    log: `logging.Logger` or `None`
        Logger used displaying logs. 

        default: `None`
    """

    def __init__(self, 
                 filename = "materials_database.db", 
                 filepath = "data", 
                 make_dirs = True,
                 key_name = "mid",
                 log = None):
        self.filename = filename
        self.filepath = filepath
        if not os.path.exists(filepath):
            if not make_dirs:
                raise FileNotFoundError("Path for storing database does not exist. To create it, set `make_dirs = True`.")
            else:
                os.makedirs(filepath)
        self._metadata = {}
        self.key_name = key_name
        self.set_logger(log)

    # properties

    @property
    def metadata(self) -> dict:
        """
        Metadata property, returns metadata attached to the backend.
        """
        return self._metadata

    @property
    def abs_path(self) -> str:
        """
        Absolute path property, contains the absolute path of the backend file.
        """
        return os.path.abspath(os.path.join(self.filepath, self.filename))

    @property
    def log(self) -> logging.Logger | None:
        """
        Logger property, returns the log.
        """
        return self._log

    # methods

    def get_single(self, mid = None, **kwargs) -> Material:
        """
        Get a single entry from the database.
        """
        raise NotImplementedError(f"Function 'get_single' is not implemented for object of class {self.__class__}")

    def get_many(self, mids = None, **kwargs) -> List[Material]:
        """
        Get a single entry from the database.
        """
        raise NotImplementedError(f"Function 'get_many' is not implemented for object of class {self.__class__}")

    def get_by_id(self, db_id) -> Material:
        """
        Return a single entry from an (integer valued) database id.
        """
        raise NotImplementedError(f"Function 'get_by_id' is not implemented for object of class {self.__class__}")

    def add_single(self, *args, **kwargs) -> None:
        """
        Add data to the database.
        """
        raise NotImplementedError(f"Function 'add_single' is not implemented for object of class {self.__class__}")
        
    def add_many(self, *args, **kwargs) -> None:
        """
        Add data to the database.
        """
        raise NotImplementedError(f"Function 'add_many' is not implemented for object of class {self.__class__}")

    def update_single(self, *args, **kwargs) -> None:
        """
        Update a single entry in the database.
        """
        raise NotImplementedError(f"Function 'update_single' is not implemented for object of class {self.__class__}")
    
    def update_many(self, *args, **kwargs) -> None:
        """
        Update several entries in the database.
        """
        raise NotImplementedError(f"Function 'update_many' is not implemented for object of class {self.__class__}")

    def update_metadata(self, *args, **kwargs) -> None:
        """
        Updata database metadata.
        """
        raise NotImplementedError(f"Function 'update_metadata' is not implemented for object of class {self.__class__}")

    def has_entry(self, entry_id) -> bool:
        """
        Check if an entry with the given id is present in the database.
        """
        raise NotImplementedError(f"Function 'has_entry' is not implemented for object of class {self.__class__}")

    def get_length(self) -> int:
        """
        Return the length of the database, i.e. the total number of entries.
        """
        raise NotImplementedError(f"Function 'get_length' is not implemented for object of class {self.__class__}")

    def set_logger(self, logger: logging.Logger) -> None:
        """
        Set logger.
        """
        self._log = logger

    # private methods

    @staticmethod
    def _log_write_message(mid):
        return f"Wrote material with id {mid}."

    def _check_select_arguments(self, mid_s, **kwargs):
        if mid_s is None and len(kwargs) == 0:
            raise ValueError("Neither material ids nor properties to select materials given.")


import numpy as np

from madas.fingerprint import Fingerprint
from madas.utils import safe_log

def DUMMY_similarity(fingerprint1: Fingerprint, 
                     fingerprint2: Fingerprint,
                     alpha = 1.0) -> np.float64:
    """
    Similarity between cartesian vectors given as DUMMYFingerprints. Calculates:

    *similarity = alpha / ( alpha + |vector1 - vector2|_2)*

    **Arguments:**

    fingerprint1, fingerprint2: *madas.fingerprints.DUMMY_fingerprint.DUMMYFingerprint*
        DUMMY fingerprints to calculate similarity.

    **Keyword arguments:**

    alpha: `float`
        Scaling factor for similarity 

        default: `1.0`

    **Returns:**
    
    similarity: *numpy.float64*
        Similarity between both fingerprints
    """
    return alpha / (alpha +np.linalg.norm(np.array(fingerprint1['data']) - np.array(fingerprint2['data'])))

def dummy_target(x):
    return sum(np.array(x)**2)

class DUMMYFingerprint(Fingerprint):
    """
    A dummy fingerprint of cartesian vectors to demonstrate the usage of learning methods.
    """

    def __init__(self, name = "DUMMY", similarity_function = DUMMY_similarity, target_function = dummy_target):
        self.set_name(name)
        self.set_fp_type("DUMMY")
        self.set_similarity_function(similarity_function)
        self.target_function = target_function

    def calculate(self, list_: list):
        """
        Generate DUMMYFingerprints from a vector.
        """
        self.set_data("data", list(list_))
        self.set_name("DUMMY_from_list")
        return self

    def from_material(self, material, property_path="", value=None):
        """
        Calculate fingerprint by setting its data property to "Material.data".
        """
        self.set_mid(material)
        if value is not None:
            self.calculate(value)
            return self
        try:
            self.set_data("data", list(material.get_data_by_path(property_path)))
        except Exception as ex_:
            if self.pass_on_exceptions:
                safe_log(f"Could not retrieve property {property_path} for material with mid {self.mid}, because of {str(ex_)}", None)
                return None 
            else:
                raise ex_
        return self

    @property
    def y(self):
        return self.target_function(self["data"])


import importlib
import json
from typing import Callable, List
from typing import Any
from copy import deepcopy
from functools import partial
import warnings

from madas import Material

from .utils import report_error, JSONNumpyEncoder

def import_builtin_module(fp_type: str, 
                          file_suffix: str = '_fingerprint', 
                          class_suffix: str = 'Fingerprint', 
                          similarity_measure_suffix: str = '_similarity'):
    """
    Function to import built-in fingerprint classes and similarity measures from different files.
    """
    module_name = 'madas.fingerprints.' + fp_type + file_suffix
    module = importlib.import_module(module_name)
    class_name = fp_type + class_suffix
    similarity_measure_name = fp_type + similarity_measure_suffix
    fingerprint_class = getattr(module, class_name)
    similarity_function = getattr(module, similarity_measure_name)
    return fingerprint_class, similarity_function

class Fingerprint():
    """
    Generic description of a fingerprint.

    Intended use case is subclassing to integrate with the `madas` framework.

    Arbitrary data can be passed to the Fingerprint object by using `Fingerprint().set_data(<key>: str, <data>: Any)`. 
    It can be retrieved either through the property `Fingerprint().data` as a dictionary, or directly via `Fingerprint()[<key>]`.

    **Parameters:**

    fp_type: `str`, `type`, or `None`
        Type of fingerprint. The behavior changes based on the type of the argument:

            - `None`: Don't intialize the fingerprint type. The object will be a generic `Fingerprint` 
            - `type`: Intialize as a fingerprint of the given type. All keyword arguments will be passed to the child class.
            - `str`: Initialize as a built-in fingerprint type using the `importfunction`. All keyword arguments will be passed to the child class.

        default: `None`

    name: `str`
        Name of the fingerprint. Can be used to distinguish between fingerprints of the same type with different parameters.

    similarity_function: `Callable` or `None`
        Function used to evaluate the similarity to other fingerprints.

        This function will be called when `Fingerprint().get_similarity()` is called.

        If set to `None` upon initialization and a fingerprint type is given, the respective
        similarity function will be set automatically.

        If set to a `callable`, *use this similarity function instead of any other*, 
        even if a fingerprint type is given.

        default: `None`

    pass_on_exceptions: `bool`
        In order to allow processing of big amounts of data without interruptions,
        the fingerprint can return 0 similarity instead of raising exceptions.
        A warning will be printed if the calculation of similarity is not successful.

        default: `False`

    importfunction: `callable`
        Function that imports fingerprint classes if the fingerprint type is provided as a string.

        default: `madas.fingerprint.import_builtin_module`

    **Methods:**
    """

    def __init__(self, fp_type: str | type | None = None, 
                       name: str | None = None,
                       similarity_function: Callable | None = None,
                       pass_on_exceptions: bool = False,
                       importfunction: Callable = import_builtin_module, 
                       **kwargs) -> None:
        if isinstance(fp_type, str) or fp_type is None:
            fingerprint_class = None
        elif isinstance(fp_type, type):
            fingerprint_class = fp_type
            fp_type = fp_type.__name__
        else:
            raise TypeError("Attibute fp_type must be either a string or a Fingerprint child type")
        name = self._name_from_type(name, fp_type)
        self._safe_set_attribute("similarity_function", similarity_function)
        self._safe_set_attribute("_fp_type", fp_type)
        self._safe_set_attribute("_name", name)
        self._safe_set_attribute("_pass_on_exceptions", pass_on_exceptions)
        self._safe_set_attribute("_data", {})
        self._safe_set_attribute("_mid", "unknown")
        if self.fp_type is not None and type(self) == Fingerprint:
            if fingerprint_class is None:
                self.set_import_function(importfunction)
                fingerprint_class, similarity_function_imported = self.importfunction(fp_type)
                similarity_function = similarity_function if similarity_function is not None else similarity_function_imported
            self.specify(fingerprint_class, **kwargs)
            #if self.similarity_function is None:
            self.set_similarity_function(similarity_function)
 
    @property
    def data(self) -> dict:
        """
        Get data dictionary of the Fingerprint.
        """
        try:
            return self._data
        except AttributeError:
            return {}

    @property
    def mid(self):
        """
        Material identifier property.
        """
        try:
            return self._mid
        except AttributeError:
            return "unknown"

    @property
    def pass_on_exceptions(self):
        """
        Property that controls if exceptions should be raised.
        """
        try:
            return self._pass_on_exceptions
        except AttributeError:
            return True

    @property
    def fp_type(self):
        """
        Property that contains the type of fingerprint.
        """
        try:
            return self._fp_type
        except AttributeError:
            return self.__class__.__name__

    @property 
    def name(self):
        """
        Property that contains the name of fingerprint.
        """
        try:
            return self._name
        except AttributeError:
            return self.fp_type

    def set_pass_on_exceptions(self, behavior: bool) -> None:
        """
        Set if fingerprint should pass and return None if an exception occures during calculation of similarities, or raise an Exception.

        **Arguments:**

        behavior: *bool*
            Set behavior.

            True -> Return *None* if exceptions are raised during calculation of similarities.

            False -> Raise exceptions.

        """
        self._pass_on_exceptions = behavior

    def set_data(self, key: str, data: Any) -> None:
        """
        Set data entries of the Fingerprint.

        **Arguments:**
        
        key: *str*
            Name of the property that is meant to be stored
        
        data: *Any*
            Properties to be stored as fingerprint data 

        """
        try:
            self._data[key] = data
        except AttributeError:
            self._data = {}
            self._data[key] = data

    def set_mid(self, mid: str | Material) -> None:
        """
        Set material id of Fingerprint.

        **Arguments:**

        mid: *str* or *Material*
            Material id to be set.

            If a Material is given, the mid will be extracted from there.

        **Returns:**

        None
        """
        if isinstance(mid, Material):
            mid = mid.mid
        self._mid = mid

    def set_fp_type(self, fp_type: str):
        if not isinstance(fp_type, str):
            raise TypeError("This function only sets the string describing the fingerprint. To specify the type of fingerprint, use the function specify().")
        self._fp_type = fp_type

    def set_name(self, name: str):
        self._name = name

    def get_similarity(self, other: object) -> float:
        """
        Calculate similarity to another fingerprint. 
        
        **Arguments:**
    
        other: *Fingerpint*
            Other fingerprint to calculate similarity

        **Returns:**

        similarity: *float*
            Similarity between both fingerprints

        **Raises:**

        `TypeError`: Different fingerprint types are used.

        **Warnings:**

        `UserWarning`: Fingerprints of different names are used.
        """
        if not self._check_type_compatibility(other):
            return 0
        if not self.name == other.name:
            warnings.warn(f"Fingerprint names {self.name} and {other.name} do not match. Use similarities with caution.")
        try:
            similarity = self.similarity_function(self, other)
            return similarity
        except Exception as exception:
            error_message = f'Could not calculate similarity for materials: {self.mid} and {other.mid} because of error: {str(exception)}'
            report_error(None, error_message)
            if self.pass_on_exceptions:
                report_error(None, "Setting similarity to 0.")
                return 0
            else:
                raise exception

    def get_similarities(self, fingerprint_list: list) -> List[float]:
        """
        Calculate similarity to all other fingerprints in *fingerprint_list*. 
        
        **Arguments:**
    
        fingerprint_list: *List[Fingerpint]*
            Other fingerprints to calculate similarity

        **Returns:**

        similarities: *List[float]*
            Similarities of self to all other fingerprints

        **Raises:**

        TypeError: Different fingerprint types are used.
        """
        similarities = [self.get_similarity(fp) for fp in fingerprint_list]
        return similarities

    def set_similarity_function(self, similarity_function: Callable, **kwargs) -> object:
        """
        Set the function for calculating similarity and return self.

        **Arguments:**
        
        similarity_function: *Callable* 
            Function that takes two fingerprint objects as input and calculates the similarity between them

        Keyword arguments are passed to the function.
        """
        if len(kwargs) != 0:
            self.similarity_function = partial(similarity_function, **kwargs)
        else:
           self.similarity_function = similarity_function
        return self

    def set_import_function(self, import_function: Callable) -> None:
        """
        Set the function used to import specialized fingerprint objects.
        
        **Arguments:**

        import_function: *Callable*; 
            A function that returns a specialized Fingerprint class and a similarity function.
        """
        self.importfunction = import_function

    def specify(self, fingerprint_class, **kwargs):
        """
        Change self to class given by ``fingerprint_class``.
        
        **Arguments:**

        fingerprint_class: *type*
            Subclass of *Fingerprint*

        Keyword arguments are passed to the __init__ function of the specialized fingerprint.
        """
        self.__class__ = fingerprint_class
        self.__init__(**kwargs)

    def calculate(self, mid: str, *args, **kwargs) -> object | None:
        """
        To be defined in child class. Calculate the fingerprint from values.
        
        **Expected return value:**
        
        self: *Fingerprint*
            It is expected that the object returns itself after calling *calculate*.
        """
        self.set_mid(mid)
        return self
    
    def from_material(self, material: Material, *args, **kwargs):
        """
        To be defined in child class. Calculate the fingerprint from a Material object.
        
        **Expected return value:**
        
        self: *Fingerprint*
            It is expected that the object returns itself after calling *calculate*.
        """
        self.set_mid(material)
        return self

    def get_data_json(self) -> str:
        """
        Return fingerprint data in as a json-coded string.
        """
        json_data = json.dumps(self.data, cls=JSONNumpyEncoder)
        return json_data
    
    def copy(self) -> object:
        """
        Return a deep copy of the fingerprint.
        """
        return deepcopy(self)

    def serialize(self) -> str:
        """
        Write fingerprint data and metadata to dictionary or string.

        **Returns:**

        data: *str*
            json dump of fingerprint
        """
        data = {
            "mid" : self.mid, 
            "name" : self.name,
            "fp_type" : self.fp_type,
            "data" : self.get_data_json(),
            "pass_on_exceptions" : self.pass_on_exceptions 
            }
        return json.dumps(data)

    @classmethod
    def deserialize(cls, data: str) -> object:
        """
        Read a fingerprint object from string data.
        This returns a *Fingerprint* object, _not_ a subclass of it. 
        The similarity function needs to be set separately, and calculate() won't work.

        **Arguments:**
        
        data: *str*
            Sting data of fingerprint (json serialized), obtained by Fingerprint().serialize().

        **Returns:**

        fingerprint: *Fingerprint*
            Fingerprint object with all fields set. 
        """
        data = json.loads(data)
        self = cls()
        self.set_mid(data["mid"])
        self.set_name(data["name"])
        self.set_fp_type(data["fp_type"])
        self.set_pass_on_exceptions(data["pass_on_exceptions"])
        self.from_data(json.loads(data["data"]))
        return self

    @staticmethod
    def _name_from_type(name, fp_type):
        if name is None:
            if isinstance(fp_type, type):
                name = fp_type.__name__
            else:
                name = str(fp_type)
        return name

    def from_data(self, data: dict): #TODO test, and make sure this is what you want
        self._data = data
        return self

    def _check_type_compatibility(self, other):
        if not isinstance(other, Fingerprint):
            error_message = f"Object of type {type(other)} is not a Fingerprint(). Can not calculate similarity"
            if self.pass_on_exceptions:
                report_error(None, error_message)
                return False
            else:
                raise TypeError(error_message)
        if not self.fp_type == other.fp_type:
            error_message = f"Can not calculate similarity for fingerprints of different types. self: {self.fp_type}, other: {other.fp_type}"
            if self.pass_on_exceptions:
                report_error(None, error_message)
                return False
            else:
                raise TypeError(error_message)
        return True

    def _safe_set_attribute(self, attibute_name: str, default: Any):
        if not hasattr(self, attibute_name):
            self.__setattr__(attibute_name, default)

    def __getitem__(self, key):
        return self.data[key]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Fingerprint):
            return False
        if self.fp_type == other.fp_type:
            if self.mid == other.mid:
                if self.name == other.name:
                    if self.get_data_json() == other.get_data_json():
                        return True
        return False

    def __repr__(self) -> str:
        return f"Fingerprint({self.fp_type}, mid = {self.mid}, name = {self.name})"



import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import mean_squared_error

from ase.visualize.plot import plot_atoms
from ase.build import make_supercell

from madas.material import Material 

def plot_minmax(target):
    plt.plot([min(target), max(target)],[min(target), max(target)],'--')

def parity_plot(pred, target, symbol = 'o', figsize = (7,7), show = True):
    plt.figure(figsize = figsize)
    plt.title('RMSE: ' + str(round(np.sqrt(mean_squared_error(pred, target)), 3)), fontsize = 35)
    max_x = max([max(pred), max(target)])*1.1
    min_x = min([min(pred), min(target)])*1.1
    plt.scatter(pred, target, marker=symbol, edgecolor = 'blue')
    plt.plot([min_x, max_x], [min_x, max_x], '--', linewidth=3, color = 'orange')
    plt.xlim(min_x, max_x)
    plt.ylim(min_x, max_x)
    plt.ylabel('Target', fontsize = 35)
    plt.xlabel('Prediction', fontsize = 35)
    plt.xticks(fontsize = 25)
    plt.yticks(fontsize = 25)
    axes = plt.gca()
    plt.setp(axes.spines.values(), linewidth=3)
    axes.xaxis.set_tick_params(width=3, length = 10)
    axes.yaxis.set_tick_params(width=3, length = 10)
    if show:
        plt.show()

def sub_numbers(string, bold = False):
    subbed_string = ''
    last_was_number = False
    for char in string:
        try:
            float(char)
            if not last_was_number:
                subbed_string += '_{'
            subbed_string += char
            last_was_number = True
        except ValueError:
            if last_was_number:
                subbed_string += '}'
            subbed_string += char
            last_was_number = False
    if last_was_number:
        subbed_string += '}'
    if bold:
        return  r'$\mathbf{'+subbed_string+r'}$'
    return  r'$\mathrm{'+subbed_string+r'}$'

def plot_material(material: Material, 
                  repeat = [2,2,2], 
                  rotation = '-66x,-2y,-6z', 
                  radii = None, 
                  show = True, 
                  show_unit_cell = 0) -> None:
    supercell = make_supercell(material.atoms, [[repeat[0],0,0], [0,repeat[1],0], [0,0,repeat[2]]])
    plot_atoms(supercell, rotation = rotation, radii = radii, show_unit_cell = show_unit_cell)
    ax = plt.gca()
    plt.setp(ax.spines.values(), visible=False) # remove outer spines
    ax.tick_params(left=False, labelleft=False)
    ax.tick_params(bottom=False, labelbottom=False)
    if show:
        plt.show()

def plot_structure_by_mid(mid, database, repeat = [3,3,1], rotation = '-66x,-2y,-6z', radii = None, show = True, return_atoms = False, show_unit_cell = 0):
    atoms = database[mid].atoms
    supercell = make_supercell(atoms, [[repeat[0],0,0], [0,repeat[1],0], [0,0,repeat[2]]])
    plot_atoms(supercell, rotation = rotation, radii = radii, show_unit_cell = show_unit_cell)
    plt.title(sub_numbers(atoms.get_chemical_formula()), fontsize = 20)
    ax = plt.gca()
    plt.setp(ax.spines.values(), visible=False) # remove outer spines
    ax.tick_params(left=False, labelleft=False)
    ax.tick_params(bottom=False, labelbottom=False)
    if show:
        plt.show()
    if return_atoms:
        return atoms



from typing import Any, List, Callable
from functools import partial
from itertools import islice
from copy import deepcopy
from multiprocessing.pool import ThreadPool
import traceback

import requests
from ase import Atoms
import numpy as np
from decouple import config as environ, UndefinedValueError

from madas.apis.api_core import APIClass, APIError
from madas import Material
from madas.utils import tqdm


def get_atoms(NOMAD_respose: dict) -> Atoms:
    """
    Creates an ASE Atoms object from a NOMAD API response.
    """
    is_optimade=False
    try:
        structure_data = NOMAD_respose["archive"]["results"]['properties']["structures"]["structure_original"]
    except KeyError:
        structure_data = NOMAD_respose["archive"]["metadata"]["optimade"]
        is_optimade=True
    pbc = NOMAD_respose["archive"]["run"][0]["system"][0]["atoms"]["periodic"]
    pos = np.array(structure_data['cartesian_site_positions'])
    cell = np.array(structure_data["lattice_vectors"])
    if not is_optimade:
        pos = pos * 1e10
        cell = cell * 1e10
    at = Atoms(positions=np.array(pos), 
               cell=np.array(cell), 
               symbols=structure_data["species_at_sites"], 
               pbc=pbc)
    return at

def get_archive(NOMAD_respose: dict) -> dict:
    """
    Return the Archive entry of a NOMAD entry from the API response. 
    """
    return NOMAD_respose['archive']

DEFAULT_PROCESSING = {
    "atoms" : get_atoms,
    "archive" : get_archive
}

DEFAULT_BASE_URL = "https://nomad-lab.eu/prod/v1/api/v1"

class API(APIClass):
    """
    API connection wrapper for the NOMAD project.

    Allows to download and process data using the NOMAD web API.

    **Keyword arguments**

    processing: `dict[str:Callable]`
        Dictionary that provides functions to process the data that is downloaded from NOMAD.
        By default, the full NOMAD Archive is stored. This is a considerable overhead if 
        not all data is required. In that case, it is suggested to limit the amout of data that is 
        requested from NOMAD (see, e.g., the docstring of `get_calculation`) and to provide a custom 
        function to process the data, such that only the necessary meta-data is kept.

        default: 

        .. code-block:: python

            DEFAULT_PROCESSING = {
                "atoms" : madas.apis.NOMAD_web_API.get_atoms,
                "archive" : madas.apis.NOMAD_web_API.get_archive
            }

    base_url: `str`
        Where to pull NOMAD data from. Only change this is you are trying
        to connect to a NOMAD OASIS instance. An example would be:
        "https://nomad-lab.eu/prod/v1/oasis/api/v1"

        To access data that requires a login, e.g., upublished data, you need to provide your credentials.
        
        This can be done by setting them as environment variables in bash:

        .. code-block:: bash 

            export NOMAD_USERNAME="MyLogin"
            export NOMAD_PASSWORD="MyPassWord"
                
        where "MyLogin" and "MyPassWord" are your NOMAD username and password, respectively.  

    timeout: `int`
        Timeout in seconds for calls to the Oasis authentication.

    force_auth: `bool`
        Force authentification if default URL is used. This can be used to access unpublished uploads on the 
        central NOMAD installation.

    logger: `logging.Logger` | `None`
        Logger to write error and info messages.

    **Methods:**         
    """

    def __init__(
            self, 
            processing=DEFAULT_PROCESSING,
            base_url: str = DEFAULT_BASE_URL,
            timeout: int = 60,
            force_auth: bool = False,
            logger=None):
        self.set_logger(logger)
        self.set_processing(processing)
        self._failed_download = set()
        self.base_url = base_url
        self.timeout = timeout
        if self.base_url != DEFAULT_BASE_URL or force_auth:
            try:
                self.headers = self._get_headers_for_authentication()
            except UndefinedValueError:
                self._report_error("Cant't find NOMAD login data in environment. Authentification not possible.")
                self.headers = None
        else:
            self.headers = None

    @property
    def processing(self):
        """
        Dictionary defining how data from NOMAD will be processed.

        Keys indicate the name of the property in MADAS.

        Functions are used to read values from the NOMAD archives.
        """
        return deepcopy(self._processing)

    @property
    def failed_download(self):
        """
        List of entry_id of calculations that could not be downloaded.

        Use `API().retry()` to try to retrieve them.
        """
        return list(self._failed_download)

    def get_calculation(self, 
                        entry_id: str, 
                        required: dict = {"required" : "*"}, 
                        return_raw: bool = False,
                        fail_quietly: bool = False) -> Material | dict:
        """
        Download data for a single calculation from NOMAD.

        **Arguments:**

        entry_id: `str`
            Unique `entry_id` as it is used by NOMAD

        **Keyword arguments:**
        
        required: `dict`
            Dictionary containing the information which sections of the NOMAD Archive are downloaded.
            This argument is passed directly to the NOMAD API and _must_ follow its definition given
            in the NOMAD documentation. Downloads the full Archive by default.

            default: ``{'required' : '*'}``

        return_raw: `bool`
            Return the response from the NOMAD API withour processing. This is helpful for debugging
            and for development of own `processing` functions.

            NOTE: The subsequent parts of the code use only the 'data' part of the response.

            default: `False`

        fail_quietly: `bool`
            Instead of raising an Exception, return a dictionary with the `entry_id`, error message and traceback.

            default: `False`

        **Returns:**

        material: `madas.material.Material`
            `Material` object with parsed properties

        or

        error_dict: `dict`
            Only if `fail_quietly==True`, dictionary describing the error that occured.

            Format: ``{error_message: str, entry_id: entry_id, traceback: traceback}`` 
        """
        url = self._URL_from_entry_id(entry_id)
        try:
            resp = requests.post(
                url, json=required, 
                headers=self.headers,
                timeout=self.timeout).json()
            if return_raw:
                return resp
            resp = resp["data"]
            atoms = self.processing["atoms"](resp) if "atoms" in self.processing.keys() else None
            data = {key:func(resp) for key, func in self.processing.items() if not key == "atoms"}
        except Exception as e:
            if fail_quietly:
                return {"error_message" : str(e), "entry_id":entry_id, "traceback" : traceback.format_exc()}
            raise e        
        return Material(entry_id, atoms=atoms, data=data)
    
    def get_calculations_by_search(self, 
                                   query: dict, 
                                   required: dict = {"required" : "*"},
                                   n_threads: int = 5,
                                   max_entries: int | None = None,
                                   ignore_entries: list | None = None) -> List[Material]:
        """
        Download several calculations from NOMAD, defined by a query. 
        Uses multithreading to maximize download speed.
        If entries can not be downloaded, a message will be written to the log.

        **Arguments:**

        query: `dict`
            Query submitted to the NOMAD API. The format must comply with the NOMAD standard.

        **Keyword arguments:**
        
        required: `dict`
            Dictionary containing the information which sections of the NONAD Archive are downloaded.
            This argument is passed directly to the NOMAD API and _must_ follow its definition given
            in the NOMAD documentation. Downloads the full Archive by default.

            default: ``{'required' : '*'}``

        n_threads: `int`
            Number of threads to start. A too high number may result in unexpected behaviour and unnecessary overhead.

            Set to any number smaller than 1 (one) to disable threading. 
            
            default: `5`


        max_entries: `int` or `None`
            Maximal number of entries to retrieve. Set to `None` to retrieve all data.

            default: `None`

        ignore_entries: `List[str]` or `None`
            Do not download entries with the any id from the given list.

            default: `None` 

        **Returns:**

        materials_list: `List[madas.material.Material]`
            List of Materials for the given query. 
                    
        """
        ids = self._query_for_entries(query, max_entries=max_entries)
        self._report_error(f"Found {len(ids)} entries", level="info")
        if max_entries is not None:
            self._report_error("Possibly not all entries discovered due to max_entries limit", level="info")        
            self._report_error(f"Downloading {int(max_entries)} entries", level="info")        
            ids = set(islice(ids, int(max_entries)))
        if ignore_entries is not None:
            for id in ignore_entries:
                ids.discard(id)
            self._report_error(f"Download data for {len(ids)} entries", level="info")
        query_function = partial(self.get_calculation, required=required, fail_quietly=True)
        if n_threads < 1:
            materials = [query_function(id_) for id_ in ids]
        else:
            with ThreadPool(int(n_threads)) as pool:
                materials = pool.map(query_function, ids)
        self._report_error("Finished download.", level="info")
        errornous_entries = []
        for idx, mat in enumerate(materials):
            if not isinstance(mat, Material):
                message = f"Could not get calculation {mat['entry_id']}, because of error {mat['error_message']} with traceback {mat['traceback']}."
                self._report_error(message)
                errornous_entries.append(idx)
        for idx in sorted(errornous_entries, reverse=True):
            failed_download = materials.pop(idx)
            self._failed_download.add(failed_download['entry_id'])
        return materials
        
    def get_property(self,
                     processing_function: Callable,
                     entry_id: str, 
                     required: dict = {"required" : "*"}) -> Any:
        """
        Receive a property from NOMAD.

        **Arguments:**

        processing_function: `Callable`
            Function to extract data from NOMAD Archive.
            For more details please refer to the documentation of the class.

        entry_id: `str`
            NOMAD entry id of the calculation that shall be downloaded.

        **Keyword arguments:**

        required: `dict`
            Dictionary containing the information which sections of the NONAD Archive are downloaded.
            This argument is passed directly to the NOMAD API and _must_ follow its definition given
            in the NOMAD documentation. Downloads the full Archive by default.

            default: ``{'required' : '*'}``

        **Returns:**

        property: `Any`
            The desired property, parsed from the NOMAD Archive.
        """
        resp = requests.post(self._URL_from_entry_id(entry_id), 
                             json=required,
                             headers=self.headers,
                             timeout=self.timeout).json()
        return processing_function(resp["data"])

    def set_processing(self, processing: dict) -> None:
        """
        Set `processing` property of the API.

        See documentation of the class for more information.
        """
        self._processing = processing

    def retry(self, required: dict = {"required" : "*"}, use_progress_bar: bool= False) -> List[Material]:
        """
        Retry previously failed downloads.

        **Keyword arguments:**

        required: `dict`
            Dictionary containing the information which sections of the NONAD Archive are downloaded.
            This argument is passed directly to the NOMAD API and _must_ follow its definition given
            in the NOMAD documentation. Downloads the full Archive by default.

            default: ``{'required' : '*'}``

        use_progress_bar: `bool`
            Show progress bar when downloading data.

        **Returns:**

        materials: `List[madas.Material]`
            Materials that could successfully be downloaded with this attempt.
        """
        if len(self.failed_download) == 0:
            self._report_error("Retry list is empty.")
            return []
        self._report_error(f"Retrying {len(self.failed_download)} entries.", level="info")
        materials = []
        for mid in tqdm(self.failed_download, disable=not use_progress_bar):
            new_mat = self.get_calculation(mid, fail_quietly=True, required=required)
            if not isinstance(new_mat, Material):
                self._report_error(f"Retry failed for material with entry_id {mid}. Error message: {new_mat['error_message']}")
                continue
            self._failed_download.discard(mid)
            materials.append(new_mat)
        return materials

    def _URL_from_entry_id(self, entry_id: str):
        return f"{self.base_url}/entries/{entry_id}/archive/query"

    def _gen_mid(self, query):
        if isinstance(query, str):
            return query
        if "entry_id:any" in query.keys():
            return query["entry_id:any"][0]
        else:
            raise ValueError("Could not retrieve entry id from query. Is this query correct?")           
            
    def _query_for_entries(self, query: dict, max_entries: int | None = None) -> List[str]:
        entries = set()
        payload = {"owner": "visible", "query": query, "required": {"include": ["entry_id"]}}
        resp = requests.post(
            f"{self.base_url}/entries/query", json=payload,
            headers=self.headers, timeout=self.timeout).json()
        entries.update(self._ids_from_response(resp))
        total_entries = self._get_max_entries(resp)
        if max_entries is not None:
            total_entries = min([total_entries, max_entries])
        while len(entries) < total_entries:
            last_len = len(entries)
            payload = self._update_pagination(payload, page_after=self._get_page_after(resp)) 
            resp = requests.post(
                f"{self.base_url}/entries/query", json=payload,
                headers=self.headers, timeout=self.timeout).json()
            entries.update(self._ids_from_response(resp))
            if not len(entries) > last_len:
                raise APIError(f"Did not add any entry ids at {len(entries)} of {total_entries} entries")
        return entries
            
    def _ids_from_response(self, response: dict) -> List[str]:
        return [entry["entry_id"] for entry in response["data"]]

    def _update_pagination(self, json_data: dict, page_after=None) -> dict:
        if page_after is not None:
            if "pagination" not in json_data.keys():
                json_data["pagination"] = {"page_after_value" : page_after}
            else:
                json_data["pagination"]["page_after_value"] = page_after
        return json_data

    def _get_page_after(self, response: dict) -> str:
        return response["pagination"]["next_page_after_value"]
    
    def _get_max_entries(self, response: dict) -> int:
        return response["pagination"]["total"]

    def _get_headers_for_authentication(self):
        """Get headers to authenticate in a NOMAD Oasis instance."""
        if not self.base_url.endswith('/api/v1'):
            self._report_error(
                f'The given URL: {self.base_url} does not appear to be a '
                f'valid NOMAD API URL, i.e., ending with /api/v1.',
                level="info")
        response = requests.get(
            self.base_url + '/auth/token',
            params={
                'username': environ('NOMAD_USERNAME'), 
                'password': environ('NOMAD_PASSWORD')
                },
            timeout=self.timeout)
        if not response.ok:
            raise ValueError(
                f'Unexpected authetentication response {response.json()}')
        token = response.json().get('access_token')
        headers = {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json'}
        return headers



from typing import Tuple
from bitarray import bitarray

class FingerprintMismatchError(Exception):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)

def match_fingerprints(fingerprint1, fingerprint2) -> Tuple[bitarray, bitarray]:
    if fingerprint1.grid_id != fingerprint2.grid_id:
        raise AssertionError(f'Can not calculate similarity of fingerprints that have been calculated with different grids.\nFingerprint 1: {fingerprint1.grid_id}\nFingerprint 2: {fingerprint2.grid_id}')
    num_bins = fingerprint1.n_state_bins
    fp1 = fingerprint1.get_bitarray()
    fp2 = fingerprint2.get_bitarray()
    start_index = max([fingerprint1.indices[0], fingerprint2.indices[0]])
    stop_index = min([fingerprint1.indices[1], fingerprint2.indices[1]])
    # find offsets
    dsp1 = (start_index - fingerprint1.indices[0]) * num_bins
    dsp2 = (start_index - fingerprint2.indices[0]) * num_bins
    dep1 = (fingerprint1.indices[1] - stop_index) * num_bins
    dep2 = (fingerprint2.indices[1] - stop_index) * num_bins
    # slice
    fp1 = fp1[dsp1:len(fp1) - dep1]
    fp2 = fp2[dsp2:len(fp2) - dep2]
    return fp1, fp2

def tanimoto_similarity(fingerprint1, fingerprint2):
    """
    Tanimoto similarity (Tc) between `DOSFingerprint` objects.

    Evaluates:

    $Tc(a, b) = a * b / ( a**2 + b**2 - a * b )$

    for binary-valued fingerprint vectors $a$ and $b$.

    Before calculating Tc, the fingerprints are matched such that the same energy 
    regions are descibed in both fingerprints.
    
    If any fingerprint is the null-vector (e.g. the energy regions do not overlap), Tc = 0.

    **Returns:**

    tc: `float`
        Tc between both fingerprints.
    """
    fp1, fp2 = match_fingerprints(fingerprint1, fingerprint2)
    if len(fp1) != len(fp2):
        raise FingerprintMismatchError(f"Can not match fingerprints of len {len(fp1)} and {len(fp2)}.")
    a = fp1.count()
    b = fp2.count()
    c = (fp1 & fp2).count()
    try:
        tc = c / float(a + b - c)
    except ZeroDivisionError:
        tc = 0
    return tc



from typing import List
import os

import numpy as np
from sklearn.cluster import DBSCAN

from madas import SimilarityMatrix

class SimilarityMatrixClusterer():
    """
    A wrapper for clustering methods to directly apply them to `SimilarityMatrix()` objects.

    **Arguments:**

    similarity_matrix: `SimilarityMatrix()` 
        Similarity matrix that should be clustered

    **Keyword arguments:**

    custerer: `type`
        a class that implements a `fit()` method that is used to cluster a np.ndarray matrix
        
        default: `sklearn.cluster.DBSCAN`
    
    clusterer_kwargs: `dict`
        Keyword arguments to be passed to the clusterer upon initializaion
        
        default: `{'metric':'precomputed', 'eps' : 0.15}`

    use_complement: `bool` 
        Switch if the similarity matrix (set to `False`) or distance matrix (set to `True`)
        is used for clustering. 
        
        Some algorithms explicitly require similarity (sometimes called affinity) matrices,
        other require distances (or, dissimilarities, metrics). To treat them all on the same 
        footing, this variable allows to set the behavior of the `matrix` property accordingly.
        
        default: `True`
    """

    def __init__(self, 
                 similarity_matrix: SimilarityMatrix, 
                 clusterer: type = DBSCAN, 
                 clusterer_kwargs: dict = {'metric':'precomputed', 'eps' : 0.15}, 
                 use_complement: bool = True):
        self.simat = similarity_matrix
        self.use_complement = use_complement
        self.clusterer = clusterer(**clusterer_kwargs)

    @property
    def matrix(self) -> np.ndarray:
        """
        Return values similarity matrix that is clustered.

        IF `self.use_complement == True`: return complement of similarity matrix
        
        ELSE: return similarity matrix
        """
        if self.use_complement:
            return 1 - self.simat.matrix
        else:
            return self.simat.matrix

    @property
    def mids(self) -> List[str]:
        """
        List of mids associated with the similarity matrix.
        """
        return [mid for mid in self.simat.mids]

    def cluster(self, **kwargs) -> object:
        """
        Perform clustering on similarity matrix and return self.

        Keyword arguments are passed to the `fit` function of the clusterer.

        **Returns:**

        self: `SimilarityMatrixClusterer`
            Self after calling `fit` of `self.clusterer`
        """
        self.clusterer.fit(self.matrix, **kwargs)
        return self

    def set_clusterer_params(self, **kwargs) -> None:
        """
        Set parameters of `self.clusterer`.
        """
        self.clusterer.set_params(**kwargs)

    @property
    def nclusters(self):
        """
        Get the number of clusters, i.e., the number of unique cluster labels.
        """
        return len(np.unique(self.labels))

    @property
    def labels(self):
        """
        Return labels of clusterer.
        """
        return self.clusterer.labels_

    def get_mids_sorted_by_cluster_labels(self, remove_orphans: bool = False):
        """
        Get mids of the similarity matrix, sorted, ascending, by cluster label.   

        **Keyword arguments:**

        remove_orphans: `bool`
            Remove all entries with cluster label -1.

            default: `False`
        """
        zip_list = list(zip(self.labels, self.mids))
        if remove_orphans:
            zip_list = filter(lambda x: x[0] != -1, zip_list)
        return [x[1] for x in sorted(zip_list, key = lambda x: x[0])]

    def save(self, 
             filename: str = 'SiMatClus.npy', 
             filepath: str = '.'):
        """
        Save clusterer to numpy format.

        **Keyword arguments:**

        filename: `str`
            Name of file to be written.

        filepath: `str`
            Relative path of file to be written.

        **Returns:**
        
        `None`
        """
        np.save(os.path.join(filepath, filename), np.array([self.simat, self.use_complement, self.clusterer], dtype=object))

    @staticmethod
    def load(filename = 'SiMatClus.npy', filepath = '.'):
        """
        Load clusterer from numpy format.

        WARNING: This allows for unpickling object files. Always make sure that the file you attempt to load is safe.

        **Keyword arguments:**

        filename: `str`
            Name of file to be loaded.

        filepath: `str`
            Relative path of file to be loaded.

        **Returns:**
        
        `self` : `SimilarityMatrixClusterer`
        """        
        simat, use_complement, clusterer = np.load(os.path.join(filepath, filename), allow_pickle = True)
        self = SimilarityMatrixClusterer(simat)
        self.clusterer = clusterer
        self.use_complement = use_complement
        return self

    def get_label_dict(self) -> dict:
        """
        Get a dictionary of all materials, where for each material id the cluster label is stored.

        **Returns:**

        label_dict: `dict`
            Dictionary {*mid1*:*label1*, [...]} mapping material ids to cluster labels.
        """        
        return dict(zip(self.mids, self.labels))

    def get_mids_by_cluster_label(self, cluster_label: int) -> List[str]:
        """
        Get mids of all materials that have the specified cluster label.

        **Arguments:**

        cluster_label: *int*
            Label of requested cluster.

        **Returns:**

        mid_list: `List[str]`
            Material ids of all materials in the specified cluster.
        """
        return [key for key, value in self.get_label_dict().items() if value == cluster_label]

    def get_cluster_sub_matrix(self, cluster_label : int) -> SimilarityMatrix:
        """
        Return a sub matrix of the similarity matrix that contains only the elements from the specified cluster.

        **Arguments:**

        cluster_label: `int`
            Label of cluster
        """
        return self.simat.get_sub_matrix(self.get_mids_by_cluster_label(cluster_label))

    def get_sorted_similarity_matrix(self):
        """
        Return a `SimilarityMatrix` where all entries are sorted by ascending cluster label.
        This is helpful for visualization.

        **Returns:**

        similarity_matrix: `SimilarityMatrix`
            Similarity matrix with sorted entries.
        """
        return self.simat.get_sub_matrix(self.get_mids_sorted_by_cluster_labels())

    @property
    def unique_labels(self) -> np.ndarray:
        """
        Get list of all unque cluster labels.

        **Returns:**

        unique_labels: `np.ndarray`
            Unique cluster labels
        """
        return np.unique(self.labels)

    def get_cluster_size(self, cluster_label: int) -> int:
        """
        Get the size (i.e. the number of members) of a specific cluster.

        **Arguments:**

        cluster_label: `int`
            Label of cluster to get size of.

        **Returns**

        cluster_size: `int`
            Number of members of the specified cluster.
        """
        return len(self.get_mids_by_cluster_label(cluster_label))



import pytest
import numpy as np

from madas import Fingerprint, SimilarityMatrix
from madas.analysis import MetricSpaceTest

def mock_mid_creator(values) -> str:
    return ".".join(str(val) for val in values)

@pytest.fixture
def fingerprints():
    points = np.arange(0,1.01,0.01), np.arange(0,1.01,0.01)
    fps = [Fingerprint("DUMMY").from_material(mock_mid_creator(point), value=list(point)) for point in zip(*points)]
    return fps

def unity(*args):
    return 1

def non_symmetric(x,y):
    if x["data"][0] <= y["data"][0]:
        return 0.5
    else:
        return 1

def non_identity(x,y):
    if x.mid == y.mid:
        return 0.8
    else:
        return 0.9

def test_init(fingerprints):

    simat = SimilarityMatrix().calculate(fingerprints)

    mst = MetricSpaceTest(fingerprints)

    assert mst.similarity_matrix == simat, "Did not calculate correct similarity matrix"

def test_call(fingerprints):

    mst = MetricSpaceTest(fingerprints)

    assert mst() == [True, True, True, True], "Test failed for metric space"

def test_uniqueness(fingerprints):

    for fp in fingerprints:
        fp.set_similarity_function(unity)

    mst = MetricSpaceTest(fingerprints)

    assert not mst(only=["uniqueness"])[0], "Uniqueness passed for non-unique space"

def test_symmetry(fingerprints):

    for fp in fingerprints:
        fp.set_similarity_function(non_symmetric)

    mst = MetricSpaceTest(fingerprints)

    assert not mst(only=["symmetry"])[0], "Symmetry passed for non-symmetric similarity measure"

def test_identity(fingerprints):

    for fp in fingerprints:
        fp.set_similarity_function(non_identity)

    mst = MetricSpaceTest(fingerprints)

    assert not mst(only=["identity"])[0], "Symmetry passed for non-identic similarity measure"

def test_triangle_inequality(fingerprints):

    mst = MetricSpaceTest(fingerprints)

    df = mst.similarity_matrix.dataframe

    mid1 = fingerprints[3].mid
    mid2 = fingerprints[4].mid

    df.loc[mid1, mid2] = 0
    df.loc[mid2, mid1] = 0

    mst.similarity_matrix.set_dataframe(df)

    assert not mst(only=["triangle inequality"])[0], "Triangle inequality passed for non-metric space"
    



import warnings
from typing import List
from itertools import combinations

from scipy.special import binom
from nomad_dos_fingerprints import DOSFingerprint, Grid
import numpy as np

from .fingerprint import Fingerprint
from .similarity import SimilarityMatrix
from .utils import tqdm


class MetricSpaceTest():
    """
    Class for fingerprints to test if the fingerprints and their respective similarity measures span a metric space.

    **Usage:**

    ``mst = MetricSpaceTest(fingerprints: List[Fingerprint])``

    ``is_metric = all(mst())``
    """
    
    def __init__(self, 
                 fingerprints: List[Fingerprint] = None, 
                 similarity_matrix: SimilarityMatrix = None, 
                 show_progress: bool = True):
        assert not (similarity_matrix is None and fingerprints is None), "Either fingerprints or similarity matrix need to be provided."   
        if similarity_matrix is None:
            self.similarity_matrix = SimilarityMatrix().calculate(fingerprints, mids = [fp.mid for fp in fingerprints], symmetric=False)
        else:
            self.similarity_matrix = similarity_matrix
        self.fingerprints = fingerprints
        self.show_progress = show_progress
        
    def __call__(self, only = None):
        only = only if only is not None else ["identity", "uniqueness", "symmetry", "triangle inequality"]
        results = [self.call_single(test_name) for test_name in only]
        return results
            
    def call_single(self, test_name : str) -> bool:
        """
        Call a single test.

        **Arguments:**

        test_name: `str`
            Name of the test. Choose from: 'uniqueness', 'identity', 'symmetry', 'triangle inequality'

        **Returns:**

        result: `bool`
            Result of the test.

        **Raises:**

        KeyError: Name of the test is not recogized.
        """
        if test_name == "uniqueness":
            result = self.uniqueness(self.similarity_matrix, self.fingerprints, show_progress = self.show_progress)
        elif test_name == "identity":
            result = self.identity(self.similarity_matrix, show_progress = self.show_progress)
        elif test_name == "symmetry":        
            result = self.symmetry(self.similarity_matrix, show_progress = self.show_progress)
        elif test_name == "triangle inequality":
            result = self.triangle_inequality(self.similarity_matrix, show_progress = self.show_progress)
        else:
            raise KeyError("No test of this name is implemented. Choose from: 'uniqueness', 'identity', 'symmetry', 'triangle inequality'")
        return result
    
    @staticmethod
    def symmetry(similarity_matrix: SimilarityMatrix, show_progress: bool = True):
        """
        Test symmetry property of a similarity matrix. The threshold for being equal is `1e-8`.

        Tests: S(i,j) == S(j,i) for fingerprints i,j and similarity S

        **Arguments:**

        similarity_matrix: `SimilarityMatrix`
            Similarity matrix to test.

        **Keyword arguments:**

        show_progress: `bool`
            Show progress of test.

            default: `True`

        **Returns:**

        result: `bool`
            Result of the test: `True`: matrix is symmetric, `False` else
        """
        if show_progress:
            print('Checking symmetry property!')
        matrix = similarity_matrix.matrix
        combs = binom(len(similarity_matrix), 2)
        symmetries = [abs(matrix[idx1][idx2] - matrix[idx2][idx1]) < 1e-8 for (idx1, idx2) in tqdm(combinations(range(len(similarity_matrix)), 2), ncols = 500, disable = not show_progress, total = combs)]
        return all(symmetries)

    
    @staticmethod
    def identity(similarity_matrix: SimilarityMatrix, show_progress: bool = True):
        """
        Test identity property of a similarity matrix.

        Tests: S(i,i) == 1 for all fingerprints i and similarity S

        **Arguments:**

        similarity_matrix: `SimilarityMatrix`
            Similarity matrix to test.

        **Keyword arguments:**

        show_progress: `bool`
            Show progress of test.

            default: `True`

        **Returns:**

        result: `bool`
            Result of the test: `True`: identical fingerprints have S == 1, `False` else
        """
        if show_progress:
            print('Checking self-similarity property!')
        matrix = similarity_matrix.matrix
        identities = [matrix[idx][idx] == 1 for idx in tqdm(range(len(similarity_matrix)), ncols = 500, disable = not show_progress)]
        return all(identities)

    @staticmethod
    def uniqueness(similarity_matrix: SimilarityMatrix, fingerprints: List[Fingerprint], show_progress: bool = True):
        """
        Test uniqueness property of a similarity matrix.

        Tests: i != j -> S(i,j) != 1 for fingerprints i,j and similarity S

        **Arguments:**

        similarity_matrix: `SimilarityMatrix`
            Similarity matrix to test.

        fingerprints: `List[Fingerprint]`
            List of fingerprints to verfy if two fingerprints are equal.

        **Keyword arguments:**

        show_progress: `bool`
            Show progress of test.

            default: `True`

        **Returns:**

        result: `bool`
            Result of the test: `True`: fingerprints are unique, `False` else
        """
        if fingerprints is None:
            raise ValueError("This test requires to give fingerprints to the MetricSpaceTest.")
        if show_progress:
            print('Checking non-identic --> S < 1 property!')
        matrix = similarity_matrix.matrix
        combs = binom(len(similarity_matrix), 2)
        for (idx1, idx2) in tqdm(combinations(range(len(similarity_matrix)), 2), ncols = 500, disable = not show_progress, total = combs):
            if idx1==idx2:
                assert False
            if matrix[idx1][idx2] == 1:
                if not fingerprints[idx1] == fingerprints[idx2]:
                    return False
        return True

    @staticmethod
    def triangle_inequality(similarity_matrix: SimilarityMatrix, show_progress: bool = True):
        """
        Test triangle inequality property of a similarity matrix. The threshold for being equal is `1e-8`.

        Tests: S(i,k) + S(j,k) <= S(j,i) + 1 for fingerprints i,j,k and similarity S

        **Arguments:**

        similarity_matrix: `SimilarityMatrix`
            Similarity matrix to test.

        **Keyword arguments:**

        show_progress: `bool`
            Show progress of test.

            default: `True`

        **Returns:**

        result: `bool`
            Result of the test: `True`: matrix obays triangle inequality, `False` else
        """
        if show_progress:
            print("Checking triangle inequality!")
        matrix = similarity_matrix.matrix
        combs = binom(len(similarity_matrix), 3)
        for (idx1, idx2, idx3) in tqdm(combinations(range(len(similarity_matrix)), 3), ncols = 500, disable = not show_progress, total = combs):
            truth = matrix[idx1][idx3] +  matrix[idx2][idx3] <=  matrix[idx1][idx2] + 1
            if not truth:
                return False
        return True

class StrideSpectrumComparison():
    """
    Compare two spectra by "striding" small spectum-fingerprint feature regions along the x-axis.
    
    **Usage**:
        For spectra 1 and 2 with x-values x1 and x2 and y-values y1 and y2:

        ssc = StrideSpectrumComparison()
        stride_list, similarities = ssc(x1, y1, x2, y2)


    **Returns:**
    
    stride_list: `list` of `float`
        List of strides that were used for calculating the similarities

    similarities:  `list` of `float`
        Similiarity of the spectra in a window centered at the respective stride
    """
        
    def __init__(self,
                 grid_id: str = 'nonuniform:-2:0.05:1.05:0.3:8:0.5:-1:1:1024',
                 cutoff: List[float] = [-1,1],
                 eref_stepsize: float = 0.1,
                 show_progress: bool = True,
                 **kwargs):
        self._grid = Grid.create(grid_id=grid_id)
        self.grid.cutoff_min, self.grid.cutoff_max = cutoff
        self.stepsize = eref_stepsize
        self.show_progress = show_progress
        self.kwargs = kwargs
    
    @property
    def grid(self):
        return self._grid
    
    def __call__(self, 
                 x1: list, 
                 y1: list, 
                 x2: list, 
                 y2: list) -> List[List[float]]:
        # determine max range of strides
        min_x = max([min(x1), min(x2)])
        max_x = min([max(x1), max(x2)])
        
        offsets = np.arange(min_x, max_x, self.stepsize)
        
        sims = []
        for offset in tqdm(offsets, disable=not self.show_progress):
            self.grid.e_ref = offset
            fp1 = DOSFingerprint(**self.kwargs).calculate(x1, y1, convert_data=None, grid_id = self.grid.get_grid_id())
            fp2 = DOSFingerprint(**self.kwargs).calculate(x2, y2, convert_data=None, grid_id = self.grid.get_grid_id())
            if any(map(lambda x: x.overflow>0, [fp1,fp2])):
                warnings.warn("Fingerprint overflows. Consider increasing Grid().n_pix.")
            if fp1.filling_factor == 0 and fp2.filling_factor == 0:
                sims.append(1)
            else:
                sims.append(fp1.get_similarity(fp2))
        return offsets.tolist(), sims
    


import pytest
from madas.fingerprint import Fingerprint
from madas.similarity import SimilarityMatrix, OverlapSimilarityMatrix, BatchedSimilarityMatrix
import os
import numpy as np
from random import random

@pytest.fixture()
def simat():
    simat = SimilarityMatrix()
    triangular_matrix = [[1,2,3],[4,5],[6]]
    mids = [str(idx) for idx in range(3)]
    simat.set_matrix(triangular_matrix)
    simat.set_mids(mids)
    return simat

class MockFingerpint():

    def __init__(self) -> None:
        self.fp_type = "Mock"
        self.name = "mock"
        self.mid = str(np.round(random(), 5))

    def calculate(self, *args, **kwargs):
        self.data =  random()
        return self

    def get_similarities(self, fps):
        value = self.data
        return [1/ (abs(value - other.data) + 1) for other in fps]

    def get_similarity(self, other):
        return 1/ (abs(self.data - other.data) + 1)

def test_SimilarityMatrix_metadata():

    fps = [MockFingerpint().calculate() for _ in range(10)]

    simat = SimilarityMatrix().calculate(fps)

    assert simat.fp_type == "Mock", "did not copy fp_type"
    assert simat.fp_name == "mock", "did not copy fp_name"

    metadata = simat.get_metadata()

    assert metadata["fp_type"] == "Mock", "did not copy fp_type to metadata"
    assert metadata["fp_name"] == "mock", "did not copy fp_name to metadata"

    simat.set_metadata({"fp_type" : "a", "fp_name" : "b"})

    assert simat.fp_type == "a", "did not set fp_type to metadata"
    assert simat.fp_name == "b", "did not set fp_name to metadata"

def test_SimilarityMatrix_set_matrix():

    simat = SimilarityMatrix()

    triangular_matrix = [[1,2,3],[4,5],[6]]

    simat.set_matrix(triangular_matrix)

    assert (simat.matrix == [[1,2,3], [2,4,5], [3,5,6]]).all(), "Did not set triangular matrix correctly"

    square_matrix = [[1,2,3], [4,5,6], [7,8,9]]

    simat.set_matrix(square_matrix)

    assert (simat.matrix == square_matrix).all(), "Did not set square matrix correctly"

def test_SimilarityMatrix_set_mids():

    simat = SimilarityMatrix()

    triangular_matrix = [[1,2,3],[4,5],[6]]

    simat.set_matrix(triangular_matrix)

    mids = [str(idx) for idx in range(3)]

    simat.set_mids(mids)

    assert all(mids == simat.mids), "Did not set mids correctly"

def test_SimilarityMatrix_set_dataframe():

    simat = SimilarityMatrix()

    square_matrix = [[1,2,3], [4,5,6], [7,8,9]]

    from pandas import DataFrame

    frame = DataFrame(square_matrix, index=["a", "b", "c"], columns=["a", "b", "c"])

    simat.set_dataframe(frame)

    assert all(simat.dataframe == frame), "Did not set dataframe correctly"

    assert all(simat.mids == ["a", "b", "c"]), "Mids were not set corretly by dataframe"

    assert (simat.matrix == square_matrix).all(), "Matrix values were not set corretly by dataframe"

def test_SimilarityMatrix_calculate():

    fps = [MockFingerpint().calculate() for _ in range(10)]

    simat = SimilarityMatrix().calculate(fps)

    assert np.allclose(simat.matrix, [fp.get_similarities(fps) for fp in fps]), "Calculated similarities are wrong"

    simat = SimilarityMatrix().calculate(fps, symmetric=False)

    assert np.allclose(simat.matrix, [fp.get_similarities(fps) for fp in fps]), "Calculated similarities (not symmetric) are wrong"

    simat = SimilarityMatrix().calculate(fps, multiprocess=None)

    assert np.allclose(simat.matrix, [fp.get_similarities(fps) for fp in fps]), "Calculated similarities (serial) are wrong"

    simat = SimilarityMatrix().calculate(fps, multiprocess=None, symmetric=False)

    assert np.allclose(simat.matrix, [fp.get_similarities(fps) for fp in fps]), "Calculated similarities (serial, not symmetric) are wrong"

def test_SimilarityMatrix_get_sub_matrix(simat):

    sub_matrix = simat.get_sub_matrix(["2", "0"])

    assert (sub_matrix.matrix == [[6,3],[3,1]]).all(), "Sub-matrix values are wrong"

    assert all(sub_matrix.mids == ["2", "0"]), "Wrong mids for sub-matrix"

    simat.get_sub_matrix(["2", "0"], copy=False)

    assert simat == sub_matrix, "Inplace sub-matrix failed"

def test_SimilarityMatrix_get_overlap_matrix(simat):

    # matrix shape:
    #
    #     | "0" | "1" | "2" 
    # --------------------
    # "0" |  1  |  2  |  3
    # "1" |  2  |  4  |  5
    # "2" |  3  |  5  |  6

    overlap_matrix = simat.get_overlap_matrix(["2"], ["1", "0"])

    assert all(overlap_matrix.row_mids == ["1", "0"]), "Wrong row mids for overlap matrix"
    assert all(overlap_matrix.column_mids == ["2"]), "Wrong row mids for overlap matrix"

    assert (overlap_matrix.matrix == [[5], [3]]).all(), "Wrong values for overlap matrix"


def test_SimilarityMatrix_lookup_similarity():

    fps = [MockFingerpint().calculate() for _ in range(10)]

    simat = SimilarityMatrix().calculate(fps)

    fp1 = Fingerprint(similarity_function=simat.lookup_similarity)
    fp1.set_mid(fps[0].mid)

    fp2 = Fingerprint()
    fp2.set_mid(fps[1].mid)

    assert fp1.get_similarity(fp1) == 1, "Wrong self-similarity for lookup similarity"

    assert fp1.get_similarity(fp2) == fps[0].get_similarity(fps[1]), "Wrong similarity for lookup similarity"

def test_SimilarityMatrix_train_test_split(simat):

    train_mids = ["2", "1"]
    test_mids = ["0"]

    train_matrix, test_matrix = simat.train_test_split(train_mids, test_mids)

    assert (train_matrix.matrix == [[6,5], [5,4]]).all(), "Wrong values for train matrix"
    assert (test_matrix.matrix == [[3, 2]]).all(), "Wrong values for test matrix"

    assert all(train_matrix.mids == train_mids), "Wrong mids for train matrix"
    assert all(test_matrix.column_mids == train_mids), "Wrong column mids for test matrix"
    assert all(test_matrix.row_mids == test_mids), "Wrong row mids for test matrix"

def test_SimilarityMatrix_align():

    mids1 = ["a", "b", "c", "d", "e"]
    matrix1 = np.array(range(1,6)) * np.eye(5)
    mids2 = ["k", "b", "c", "j", "e"]
    matrix2 = np.array(range(6,11)) * np.eye(5)
    mids3 = ["b", "c"]
    matrix3 = np.array([11,12]) * np.eye(2)

    simat1 = SimilarityMatrix(matrix = matrix1, mids = mids1)
    simat2 = SimilarityMatrix(matrix = matrix2, mids = mids2)

    simat1.align(simat2)

    assert all(simat1.mids == ["b", "c", "e"]), "Did not align mids matrix1"
    assert all(simat2.mids == ["b", "c", "e"]), "Did not align mids matrix2"

    assert (simat1.matrix == [[2,0,0], [0,3,0], [0,0,5]]).all(), "Did not align values matrix1"
    assert (simat2.matrix == [[7,0,0], [0,8,0], [0,0,10]]).all(), "Did not align values matrix2"

    simat1 = SimilarityMatrix(matrix = matrix1, mids = mids1)
    simat2 = SimilarityMatrix(matrix = matrix2, mids = mids2)
    simat3 = SimilarityMatrix(matrix = matrix3, mids = mids3)

    simat1.align([simat2, simat3])

    assert all(simat1.mids == ["b", "c"]), "Did not align mids matrix1"
    assert all(simat2.mids == ["b", "c"]), "Did not align mids matrix2"
    assert all(simat3.mids == ["b", "c"]), "Did not align mids matrix3"

    assert (simat1.matrix == [[2,0], [0,3]]).all(), "Did not align values matrix1"
    assert (simat2.matrix == [[7,0], [0,8]]).all(), "Did not align values matrix2"
    assert (simat3.matrix == [[11,0], [0,12]]).all(), "Did not align values matrix3"

def test_SimilarityMatrix_get_entry(simat):

    assert simat.get_entry("0", "2") == 3, "Did not get correct value"

def test_SimilarityMatrix_get_row(simat):

    assert all(simat.get_row("1") == [2,4,5]), "Got wrong row values"

def test_SimilarityMatrix_get_unique_entries(simat):

    assert (simat.get_unique_entries() == list(range(1,7))).all(), "Wrong unique entries"

def test_SimilarityMatrix_get_k_most_similar(simat):

    most_similar = simat.get_k_most_similar("2")

    assert most_similar == {"1" : 5.0, "0" : 3.0}, "Wrong list of most similar materials"

def test_SimilarityMatrix_save_load(simat, tmpdir):

    simat.save(filepath = str(tmpdir))

    matrix = SimilarityMatrix.load(filepath=str(tmpdir))

    assert simat == matrix, "Save/load did not work properly"

    with pytest.raises(FileExistsError):
        simat.save(filepath = str(tmpdir))

def test_SimilarityMatrix_get_matching_matrices():

    mids1 = ["a", "b", "c", "d", "e"]
    matrix1 = np.array(range(1,6)) * np.eye(5)
    mids2 = ["k", "b", "c", "j", "e"]
    matrix2 = np.array(range(6,11)) * np.eye(5)

    simat1 = SimilarityMatrix(matrix = matrix1, mids = mids1)
    simat2 = SimilarityMatrix(matrix = matrix2, mids = mids2)

    simat1a, simat2a = simat1.get_matching_matrices(simat2)

    assert all(simat1a.mids == ["b", "c", "e"]), "Did not match mids matrix1"
    assert all(simat2a.mids == ["b", "c", "e"]), "Did not match mids matrix2"

    assert (simat1a.matrix == [[2,0,0], [0,3,0], [0,0,5]]).all(), "Did not match values matrix1"
    assert (simat2a.matrix == [[7,0,0], [0,8,0], [0,0,10]]).all(), "Did not match values matrix2"

def test_SimilarityMatrix_get_cleared_matrix(simat):

    cleared = simat.get_cleared_matrix(["1"])

    assert all(cleared.mids == ["0", "2"]), "Did not remove mid"

    assert (cleared.matrix == [[1,3], [3,6]]).all(), "Did not remove entries"

def test_SimilarityMatrix_add(simat):

    from numpy.random import shuffle, seed
    from copy import deepcopy

    seed(0)
    mids = simat.mids
    shuffled_mids = deepcopy(mids)
    shuffle(shuffled_mids)

    assert shuffled_mids.tolist() != mids.tolist(), "Shuffling did not work"

    simat2 = simat.get_sub_matrix(shuffled_mids)
    
    assert ((simat2 + simat).matrix == (simat * 2).matrix).all()
    assert (simat2 * 0.5 + simat * 0.5) == simat

def test_SimilarityMatrix_sub(simat):

    from numpy.random import shuffle, seed
    from copy import deepcopy

    seed(0)
    mids = simat.mids
    shuffled_mids = deepcopy(mids)
    shuffle(shuffled_mids)

    assert shuffled_mids.tolist() != mids.tolist(), "Shuffling did not work"

    simat2 = simat.get_sub_matrix(shuffled_mids)
    
    assert (simat2 * 2 - simat) == simat

def test_OverlapSimilarityMatrix_set_mids():

    matrix = OverlapSimilarityMatrix([[1,0,0],[0,0,1]])

    matrix.set_mids(["a","b"], ["c", "d", "e"])

    rmids, cmids = matrix.mids

    assert all(rmids == ["a","b"]), "Did not set row mids correctly"
    assert all(cmids == ["c", "d", "e"]), "Did not set row mids correctly"

def test_OverlapSimilarityMatrix_calculate():

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    assert matrix.fp_type == "Mock", "Did not set fp_type"
    assert matrix.fp_name == "mock", "Did not set fp_name"

    mat = [fp.get_similarities(fps[:10]) for fp in fps[10:]]

    assert (mat == matrix.matrix).all(), "Did not calculate correct matrix"

    rmids, cmids = matrix.mids

    assert all(cmids == [fp.mid for fp in fps[:10]]), "Did not set correct column mids"
    assert all(rmids == [fp.mid for fp in fps[10:]]), "Did not set correct row mids"

def test_OverlapSimilarityMatrix_get_entries():

    matrix = OverlapSimilarityMatrix([[1,2,3], [4,5,6], [7,8,9]])

    assert all(matrix.get_entries() == list(range(1,10))), "Did not return correct entries"

def test_OverlapSimilarityMatrix_get_row():

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    row = matrix.get_row(0)

    assert all(row == fps[10].get_similarities(fps[:10])), "Did not return correct row by index"

    row = matrix.get_row(fps[11].mid)

    assert all(row == fps[11].get_similarities(fps[:10])), "Did not return correct row by mid"


def test_OverlapSimilarityMatrix_get_column():

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    column = matrix.get_column(0)

    assert all(column == fps[0].get_similarities(fps[10:])), "Did not return correct column by index"

    column = matrix.get_column(fps[1].mid)

    assert all(column == fps[1].get_similarities(fps[10:])), "Did not return correct column by mid"

def test_OverlapSimilarityMatrix_dataframe():

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    df = matrix.dataframe

    from pandas import DataFrame

    df2 = DataFrame(matrix.matrix, index = [fp.mid for fp in fps[10:]], columns=[fp.mid for fp in fps[:10]])

    assert all((df == df2).all()), "Did not return correct dataframe"

def test_OverlapSimilarityMatrix_get_sub_matrix():

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    sub_row_mids = [fp.mid for fp in fps[12:14]]
    sub_column_mids = [fp.mid for fp in fps[2:8]]

    sub_matrix = matrix.get_sub_matrix(sub_row_mids, sub_column_mids)

    sub_matrix_values = [fp.get_similarities(fps[2:8]) for fp in fps[12:14]]

    assert (sub_matrix.matrix == sub_matrix_values).all(), "Wrong values for sub matrix"

    assert all(sub_matrix.column_mids == sub_column_mids), "Wrong column mids for sub matrix"

    assert all(sub_matrix.row_mids == sub_row_mids), "Wrong row mids for sub matrix"

    matrix.get_sub_matrix(sub_row_mids, sub_column_mids, copy = False)

    assert matrix == sub_matrix, "Not copying does not return correct sub matrix"

def test_OverlapSimilarityMatrix_save_load(tmpdir):

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    matrix.save(filepath = str(tmpdir))

    loaded = SimilarityMatrix.load(filename="overlap_similarity_matrix.npy", filepath=str(tmpdir))

    assert matrix == loaded, "Did not save/load correct matrix data"

    with pytest.raises(FileExistsError):
        matrix.save(filepath = str(tmpdir))


def test_OverlapSimilarityMatrix_get_entry():

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    row_mid = fps[13].mid
    column_mid = fps[2].mid

    sim = fps[2].get_similarity(fps[13])

    assert matrix.get_entry(row_mid, column_mid) == sim, "Returned wrong similarity value"

def test_OverlapSimilarityMatrix_transpose():

    fps = [MockFingerpint().calculate() for _ in range(15)]

    matrix = OverlapSimilarityMatrix().calculate(fps[:10], fps[10:])

    row_mids, column_mids = matrix.mids

    matrix.transpose()

    assert all(matrix.row_mids == column_mids), "Did not transpose row mids"
    assert all(matrix.column_mids == row_mids), "Did not transpose column mids"

    assert (matrix.matrix == [fp.get_similarities(fps[10:]) for fp in fps[:10]]).all(), "Does not contain correct transposed similarities"

def test_OverlapSimilarityMatrix_align():

    r_mids1 = ["a", "b", "c"]
    c_mids1 = ["e", "f"]

    r_mids2 = ["g", "b", "c"]
    c_mids2 = ["e", "k"]

    matrix1 = [
        [1, 0],
        [0, 1],
        [1, 1]
    ]

    matrix2 = 2 * np.array(matrix1)

    osm1 = OverlapSimilarityMatrix(matrix1, r_mids1, c_mids1)
    osm2 = OverlapSimilarityMatrix(matrix2, r_mids2, c_mids2)

    osm1.align(osm2)

    assert (osm1.matrix == [[0], [1]]).all(), "Did not align matrix1 values correctly"
    assert (osm2.matrix == [[0], [2]]).all(), "Did not align matrix2 values correctly"

    assert all(osm1.row_mids == ["b", "c"]), "Did not align row mids matrix1"
    assert all(osm2.row_mids == ["b", "c"]), "Did not align row mids matrix2"
    assert all(osm1.column_mids == ["e"]), "Did not align column mids matrix1"
    assert all(osm2.column_mids == ["e"]), "Did not align column mids matrix2"

    r_mids1 = ["a", "b", "c"]
    c_mids1 = ["e", "f"]

    r_mids2 = ["g", "b", "c"]
    c_mids2 = ["e", "k"]

    r_mids3 = ["g", "j", "c"]
    c_mids3 = ["e", "k"]

    matrix1 = [
        [1, 0],
        [0, 1],
        [1, 1]
    ]

    matrix2 = 2 * np.array(matrix1)
    matrix3 = 3 * np.array(matrix1)

    osm1 = OverlapSimilarityMatrix(matrix1, r_mids1, c_mids1)
    osm2 = OverlapSimilarityMatrix(matrix2, r_mids2, c_mids2)
    osm3 = OverlapSimilarityMatrix(matrix3, r_mids3, c_mids3)

    osm1.align([osm2, osm3])

    assert (osm1.matrix == [[1]]).all(), "Did not align matrix1 values correctly"
    assert (osm2.matrix == [[2]]).all(), "Did not align matrix2 values correctly"
    assert (osm3.matrix == [[3]]).all(), "Did not align matrix3 values correctly"

    assert all(osm1.row_mids == ["c"]), "Did not align row mids matrix1"
    assert all(osm2.row_mids == ["c"]), "Did not align row mids matrix2"
    assert all(osm3.row_mids == ["c"]), "Did not align row mids matrix3"
    assert all(osm1.column_mids == ["e"]), "Did not align column mids matrix1"
    assert all(osm2.column_mids == ["e"]), "Did not align column mids matrix2"
    assert all(osm3.column_mids == ["e"]), "Did not align column mids matrix3"

def test_BatchedSimilarityMatrix_init_from_metadata(tmpdir):

    bsm = BatchedSimilarityMatrix(root_path=tmpdir, size = 10)

    bsm.save_metadata()

    del bsm

    bsm = BatchedSimilarityMatrix(root_path=tmpdir)

    assert bsm.size == 10, "Did not read from metadata file"

def test_BatchedSimilarityMatrix_fingerprint_file_batches(tmpdir):
    
    from madas.fingerprints.DUMMY_fingerprint import DUMMYFingerprint
    from itertools import product
    import json

    fps = [DUMMYFingerprint().from_material(".".join(map(str,prod)), value=prod) for prod in product(range(5), repeat = 2)]

    fp_data_serialized = [fp.serialize() for fp in fps]

    with open(os.path.join(tmpdir, "batched_fingerprints_test_full_list.json"), "w") as f:
       json.dump(fp_data_serialized, f)

    bsm = BatchedSimilarityMatrix(root_path=tmpdir, batch_size = 5)

    bsm.fingerprint_file_batches("batched_fingerprints_test_full_list.json", tmpdir, overwrite=True)

    del bsm

    bsm = BatchedSimilarityMatrix(root_path=tmpdir)

    assert bsm.size == 25, "Did not save correct size to metadata"

    fp_data = []

    for batch in bsm.batch_iterator.linear_batch_list(bsm.size, bsm.batch_size):
        fp_file_name = bsm.gen_fingerprint_batch_file_name(batch[0], batch[1])
        fp_file_path = os.path.join(bsm.folder_path, fp_file_name)
        assert os.path.exists(fp_file_path), f"Fingerprint file for batch {batch} not found at {fp_file_path}"
        with open(fp_file_path, "r") as f:
            data = json.load(f)
        deserialized_data = [Fingerprint.deserialize(entry) for entry in data]
        assert bsm.fingerprints_from_batch_file(batch[0], batch[1]) == deserialized_data, "Fingerprints from batch file did not return correct data"
        fp_data.extend(data)

    assert fp_data == fp_data_serialized, "Could not recover fingerprint data from files"

    mid_file_path = os.path.join(bsm.folder_path, bsm.mid_filename)

    with open(mid_file_path, "r") as f:
        mids = [line.split()[1] for line in f.readlines()]

    assert mids == [fp.mid for fp in fps], "Did not write correct mids to mid file"

def test_BatchedSimilarityMatrix_calculate_and_retrieve_results(tmpdir):

    from madas.fingerprints.DUMMY_fingerprint import DUMMYFingerprint, DUMMY_similarity
    from itertools import product
    import json

    fps = [DUMMYFingerprint().from_material(".".join(map(str,prod)), value=prod) for prod in product(range(5), repeat = 2)]

    ref_simat = SimilarityMatrix(dtype=np.float32).calculate(fps)

    fp_data_serialized = [fp.serialize() for fp in fps]

    with open(os.path.join(tmpdir, "batched_fingerprints_test_full_list.json"), "w") as f:
       json.dump(fp_data_serialized, f)

    bsm_t1 = BatchedSimilarityMatrix(root_path=tmpdir, batch_size = 10, n_tasks=2, task_id=0)
    bsm_t1.fingerprint_file_batches("batched_fingerprints_test_full_list.json", tmpdir)
    bsm_t2 = BatchedSimilarityMatrix(root_path=tmpdir, batch_size = 10, n_tasks=2, task_id=1)

    assert bsm_t1.matrices_for_this_task == 3, "Wrong number of matrices for first task"
    assert bsm_t2.matrices_for_this_task == 3, "Wrong number of matrices for second task"

    bsm_t1.calculate(DUMMY_similarity)

    for batch in bsm_t1.batch_iterator:
        file_name = bsm_t1.gen_matrix_batch_file_name(batch)
        file_path = os.path.join(bsm_t1.folder_path, file_name)
        assert os.path.exists(file_path), "Did not calculate similarity matrix batch task 0"

    for batch in bsm_t2.batch_iterator:
        file_name = bsm_t2.gen_matrix_batch_file_name(batch)
        file_path = os.path.join(bsm_t2.folder_path, file_name)
        assert not os.path.exists(file_path), "Did calculate wrong similarity matrix batch"

    bsm_t2.calculate(DUMMY_similarity)
    
    for batch in bsm_t2.batch_iterator:
        file_name = bsm_t2.gen_matrix_batch_file_name(batch)
        file_path = os.path.join(bsm_t2.folder_path, file_name)
        assert os.path.exists(file_path), "Did not calculate similarity matrix batch task 1"

    batched_matrix_values = []
    for mid in ref_simat.mids:
        row = bsm_t1.get_row_by_mid(mid)
        batched_matrix_values.append(row)

    assert isinstance(batched_matrix_values[0][0], np.float32), "Wrong data type for matrix elements"

    batched_matrix_values = np.array(batched_matrix_values, dtype=np.float32)

    assert (ref_simat.matrix == batched_matrix_values).all(), "Did not reproduce reference similarity matrix"

    bsm_t1.write_most_similar_materials_file(k = 10)
    bsm_t2.write_most_similar_materials_file(k = 10)

    msm_t1_path = os.path.join(bsm_t1.folder_path, bsm_t1.most_similar_materials_filename)
    msm_t2_path = os.path.join(bsm_t2.folder_path, bsm_t2.most_similar_materials_filename)

    assert os.path.exists(msm_t1_path), "Did not write most similar materials file task 0"
    assert os.path.exists(msm_t2_path), "Did not write most similar materials file task 1"

    msm_bsm = []

    with open(msm_t1_path, 'r') as f:
        for line in f:
            msm_bsm.append(json.loads(line))
    with open(msm_t2_path, 'r') as f:
        for line in f:
            msm_bsm.append(json.loads(line))

    ref_msm = [{mid : ref_simat.get_k_most_similar(mid, k=10)} for mid in ref_simat.mids]

    msm_bsm.sort(key = lambda x: [idx for idx, entry in enumerate(ref_msm) if list(entry)[0] == list(x)[0]][0])

    assert len(msm_bsm) == len(ref_msm), "Inconsistent number of most similar materials"

    for r1, r2 in zip(msm_bsm, ref_msm):
        ref_mid1 = list(r1)[0]
        ref_mid2 = list(r2)[0]
        assert ref_mid1 == ref_mid2, "Wrong mids in line!"
        data1 = r1[ref_mid1]
        data2 = r2[ref_mid2]
        for key in data2.keys():
            if key not in data1.keys():
                other_key = [key1 for key1 in data1.keys() if key1 not in data2.keys()][0]
                assert ref_simat.get_entry(ref_mid1, key) == ref_simat.get_entry(ref_mid1, other_key), "Similiarties for ambiguous entries are not identical."



import os
import json
import random
import logging
from typing import Callable, List
from typing import Any
from traceback import format_exc as get_tb_string

import numpy as np
import pandas as pd

from .fingerprint import Fingerprint
from .similarity import SimilarityMatrix
from .material import Material
from .utils import tqdm

class MaterialsDatabase():
    """
    A database wrapper to simplify materials data download from online repositories and study the similarity of materials based on different measures.
    Materials in the database can be accessed via material identifiers (mid).

    **Keyword arguments**

    filename: *str*
        Name of database file. Ignored if a Backend is specified explicitly.

        default: 'materials_database.db'

    filepath: *str*
        Path of database file. Ignored if a Backend is specified explicitly.

        default: 'data'

    key_name: *str*
        Name of unique key used in the database backend. Ignored if a Backend is specified explicitly.

        default: 'mid'

    api: *madas.apis.api_core.APIClass* object or *None*
        API object that provides an interface to web databases
        Default will use the NOMAD Encyclopedia API.

        default: None

    backend: *str* or *madas.backend.backend_core.Backend* object
        Name of database backend to use or backend object.
        Default is ASEs AtomsDatabase.

        default: 'ase'

    log_mode: *str*
        Logging mode: choose between:

        `"full"`: Write to screen and log file

        `"silent"` : Write to file only

        `"stream"` : Write to screen only

        `"None"` : Do not log

        default: `"full"`

    **Methods:**
    """

    def __init__(self, 
                 filename: str = 'materials_database.db', 
                 filepath: str = 'data',
                 name: str | None = None,
                 key_name: str = 'mid',
                 api: object | None = None, 
                 backend: str | object = 'ase', 
                 log_mode: str = "full"):
        # initialize logs
        if  log_mode is not None and log_mode.lower() != "none":
            self._init_loggers(filepath, filename.split('.db')[0], log_mode = log_mode)
        else:
            self.log, self.api_logger = None, None

        # initialization of database backend
        from madas.backend import Backend
        if backend == 'ase':
            from madas.backend import ASEBackend
            self.backend = ASEBackend(filename, filepath, key_name = key_name, log = self.api_logger)
        elif isinstance(backend, Backend):
            self.backend = backend
            self.backend.set_logger(self.api_logger)
        else:
            raise AttributeError("Ivalid argument for backend: Choose from ['ase'] or pass a Backend object.")

        # initialization of API
        if api is None:
            from .apis.NOMAD_web_API import API
            self.api = API(logger = self.api_logger)
        else:
            self.api = api
            self.api.set_logger(self.api_logger)

        # import api doc strings
        self.add_material.__func__.__doc__ += f"\n{self.api.get_calculation.__func__.__doc__}" 
        self.fill_database.__func__.__doc__ += f"\n{self.api.get_calculations_by_search.__func__.__doc__}"
        self.add_property.__func__.__doc__ += f"\n{self.api.get_property.__func__.__doc__}"

        # miscellaneous
        self._iter_index = 0
        self.query_encoder = json.JSONEncoder
        self.set_name(name)

    @property
    def log_file_path(self):
        handlers = list(filter(lambda x: x.get_name().split('_')[-1] == 'error', self.log.handlers))
        return handlers[0].baseFilename

    @property
    def name(self):
        if self._name is None:
            if "database_name" in self.get_metadata().keys():
                self._name = self.get_metadata()["database_name"]
            else:
                return "unnamed"
        return self._name
    
    def set_name(self, name: str | None):
        """
        Set name of the database.
        """
        self._name = str(name) if name is not None else None
        self._update_metadata({"database_name":self.name})

    def get_property(self, 
                     mid: str, 
                     property_name: str) -> Any:
        """
        Get a property of a single material specified by name or path in Material object from the database.

        **Arguments:**

        mid: *string*
            Material id of the requested material
        
        property_name: *str*
            Name or path in Material object of the requested property

            Example:

            "code_name" --> stored in Material().data["code_name"] retrieves the code used for the DFT calculation

            "electronic_dos/dos_values" --> Material().data["electronic_dos"]["dos_values"] retrieves the DOS values of the material

        **Returns:**
        
        property: *Any* --> property value if it exists in the database

        *None* --> else
        """
        material = self.backend.get_single(mid)
        prop = self._property_from_material(material, property_name)
        return prop

    def get_properties(self, 
                       property_name: str, 
                       output_mids: bool = False,
                       show_progress: bool = True) -> List[Any]:
        """
        Get a list of properties for all materials of the database.
        
        **Arguments:**
        
        property_name: *str*
            Name or path in Material object of the requested property

            Example:

            "code_name" --> stored in Material().data["code_name"] retrieves the code used for the DFT calculation

            "electronic_dos/dos_values" --> Material().data["electronic_dos"]["dos_values"] retrieves the DOS values of the material

        **Keyword arguments:**
        
        output_mids: *bool*
            Output the list of corresponding material ids together with the properties

            default: False

        show_progress: `bool`
            Show a progress bar.

            default: `True`

        **Returns:**
        
        tuple of:

        properties: *list*
            List of all properties of materials. If a property can not be retrieved, it will be set to *None*
        
        ,
        
        [mids]: *list*
            List of material ids
        """
        properties = []
        mids = []
        for material in tqdm(self, disable = not show_progress):
            prop = self._property_from_material(material, property_name)
            properties.append(prop)
            mids.append(material.mid)
        if output_mids:
            return properties, mids
        return properties

    def get_property_dataframe(self, property_paths: List[str]) -> pd.DataFrame:
        """
        Generate a pandas DataFrame object that contains a table of properties.

        **Arguments:**

        property_paths: *List[str]*
            List of property paths that point to the respective property.

        **Returns:**

        property_dataframe: *pandas.DataFrame*
            Dataframe, where the index contains the mids and the columns the respective properties.
        """
        _data = {}
        for material in self:
            new_entry = {}
            for prop_name in property_paths:
                new_entry[prop_name] = self._property_from_material(material, prop_name)
            _data[material.mid] = new_entry
        return pd.DataFrame(data=_data).T


    def get_fingerprint(self, 
                        fp_type: str, 
                        mid: str, 
                        name: str = None,
                        force_calculate = False,
                        similarity_function: Callable = None,
                        fingerprint_kwargs: dict = {}, **kwargs) -> Fingerprint:
        """
        Get a single fingerprint of type *fp_type* for a material with id *mid*.

        **Arguments:**

        fp_type: *str*
            Type of fingerprint X, must correspond to a XFingerprint() object defined elsewhere

        mid: *str*
            Material id of the requested material

        **Keyword arguments:**

        name: *str*
            Name of the fingerprint as used in the database, if name == None: name = fp_type. 
            This parameter is used to distinguish between fingerprints in the database, i.e. it must be unique for each unique fingerprint.
            
            default: None

        force_calculate: *bool*
            Force calculation of the fingerprint even if fingerprint data was stored in the database before.

            default: False

        similarity_function: *Callable*
            Similarity function that takes two fingerprints as arguments and calculates their similarity.

            default: None

        fingerprint_kwargs: *dict*
            Additional keyword arguments that are passed to Fingerprint().__init__.

            default: {}

        Additional keyword arguments are passed to Fingerprint().calculate().        

        **Returns:**
        
        Fingerprint() object

        None if calculation of Fingerprint failed

        **Raises:**

        KeyError: No material with specified id in the database.
        """
        material = self[mid]
        fingerprint = self._get_fingerprint(material, fp_type, name, fingerprint_kwargs, force_calculate, **kwargs)
        if similarity_function is not None:
            fingerprint.set_similarity_function(similarity_function)
        return fingerprint

    def get_fingerprints(self, 
                         fp_type: str | type, 
                         name: str = None, 
                         fingerprint_kwargs: dict = {},
                         force_calculate = False, 
                         similarity_function: Callable = None,
                         show_progress: bool = True, **kwargs) -> List[Fingerprint]:
        """
        Get fingerprints of type *fp_type* for all materials in the database.

        Generates fingerprints if they don't exist. To retrieve existing fingerprints, 

                **Arguments:**

        fp_type: *str* or `type`
            Type of fingerprint X, must correspond to a XFingerprint() object defined elsewhere

        **Keyword arguments:**

        name: *str*
            Name of the fingerprint as used in the database, if name == None: name = fp_type. 
            This parameter is used to distinguish between fingerprints in the database, i.e. it must be unique for each unique fingerprint.
            
            default: None

        force_calculate: *bool*
            Force calculation of the fingerprint even if fingerprint data was stored in the database before.

            default: False

        fingerprint_kwargs: *dict*
            Additional keyword arguments that are passed to Fingerprint().__init__.

            default: {}

        similarity_function: *Callable*
            Similarity function that takes two fingerprints as arguments and calculates their similarity.

            default: None

        show_progress: *bool*
            Show a progress bar during calculation.

            default: True

        Additional keyword arguments are passed to Fingerprint().calculate().        

        **Returns:**
        
        List[Fingerprint() object or None]
        """
        fingerprints = []
        name = Fingerprint._name_from_type(name, fp_type)
        if name in self.get_metadata().get('fingerprints', []):
            if force_calculate:
                self.log.info(f'Fingerprints with name "{name}" exist in database. Generating new.')
            else:
                self.log.info(f'Fingerprints with name "{name}" exist. Reading from database.')
        else:
            self.log.info(f'Generating "{fp_type}" fingerprints.')
        for material in tqdm(self, disable = not show_progress):
            fp = self._get_fingerprint(material, fp_type, name, fingerprint_kwargs, force_calculate, **kwargs)
            if similarity_function is not None:
                fp.set_similarity_function(similarity_function)
            fingerprints.append(fp) 
        return fingerprints

    def get_similarity_matrix(self, fp_type: str | type, name: str = None, dtype = np.float64, **kwargs) -> SimilarityMatrix:
        """
        Calculate a SimilarityMatrix() object from Fingerprints of type *fp_type* from all entries of the database.
        If fingerprints for some entries can not be calculated, they will be excluded from the database.

        **Arguments:**

        fp_type: *str* or `type`
            Type of fingerprint X, must correspond to a XFingerprint() object defined elsewhere

        **Keyword arguments:**

        name: *str*
            Name of the fingerprint as used in the database, if name == None: name = fp_type. 
            This parameter is used to distinguish between fingerprints in the database, i.e. it must be unique for each unique fingerprint.
            
            default: None

        dtype: *type*
            Data type used by SimilarityMatrix to store similarities

            default: numpy.float64

        Additional keyword arguments are passed to SimilarityMatrix().calculate().
                
        **Returns:**

        SimilarityMatrix() object

        **Raises:**

        ValueError: Fingerprints could not be obtained, no matrix can be calculated
        """
        fps = self.get_fingerprints(fp_type, name = name)
        fps = [fp for fp in fps if fp is not None]
        if len(fps) < 1:
            self.log.error('No fingerprints loaded (not generated?).')
            raise ValueError('No fingerprints fitting criteria.')
        simat = SimilarityMatrix(dtype = dtype).calculate(fps, **kwargs)
        return simat

    def add_fingerprint(self, 
                        fp_type: str | type, 
                        name: str = None, 
                        show_progress: bool = True, 
                        force_calculate: bool = False, 
                        fingerprint_kwargs: dict = {}, **kwargs) -> None:
        """
        Calculate fingerprints of all materials in the database and store them.

        **Arguments:**
        
        fp_type: *str* or *type*
            Type of fingerprint X, must correspond to a XFingerprint() object

            Fingerprint types can also be added as `type`s, then the data will be stored using
            the function `Fingerprint().serialize()`. Deserialized fingerprints from the database
            will be generic `Fingerprint` objects and the similarity function is not set.

        **Keyword arguments:**

        name: *str*
            Name of the fingerprint as used in the database, if name == None: name = fp_type. 
            This parameter is used to distinguish between fingerprints in the database, i.e. it must be unique for each unique fingerprint.
            
            default: None

        force_calculate: *bool*
            Force calculation of the fingerprint even if fingerprint data was stored in the database before.

            default: False

        fingerprint_kwargs: *dict*
            Additional keyword arguments that are passed to Fingerprint().__init__.

            default: {}

        show_progress: *bool*
            Show a progress bar during calculation.

            default: True

        Additional keyword arguments are passed to Fingerprint().calculate().        
        """
        name = Fingerprint._name_from_type(name, fp_type)
        if name in self.get_metadata().get('fingerprints', []):
            if force_calculate:
                self.log.info(f'Fingerprints with name "{name}" exist. Overwriting...')
            else:
                self.log.info(f'Fingerprints with name "{name}" exist. To overwrite, set `force_calculate=True`.')
                return
        self.log.info(f'Generating {name} fingerprints...')
        fingerprints = self.get_fingerprints(fp_type, name = name, fingerprint_kwargs = fingerprint_kwargs, force_calculate = force_calculate, show_progress = show_progress, **kwargs)
        self.log.info(f'Writing {name} fingerprints to database...')
        # remove entries where calculation failed 
        fingerprints = list(filter(lambda x: x is not None, fingerprints))
        mid_list = [fingerprint.mid for fingerprint in fingerprints]
        if isinstance(fp_type, type):
            dictionary_list = [{fingerprint.name : fingerprint.serialize()} for fingerprint in fingerprints]
        else:
            dictionary_list = [{fingerprint.name : fingerprint.get_data_json()} for fingerprint in fingerprints]
        self.update_entries(mid_list, dictionary_list)
        self.log.info(f'Finished generation of {name} fingerprints.')
        self._update_metadata({'fingerprints' : [name]}) #TODO Adapt new metadata schema

    def add_fingerprints(self, 
                         fp_types: List[str | type], 
                         names: List[str] = [None], 
                         show_progress = False,
                         fingerprint_kwargs_list: List[dict] = None,
                         fingerprint_calculate_kwargs_list: List[dict] = None,
                         force_calculate = False):
        """
        Calculate several fingerprints of each material in the database and store them.
        Because storing data can be slow for large database, this method should be preferred over adding fingerprints one by one.

        **Arguments:**
        
        fp_types: *List[str]*
            List of types of fingerprint X, must correspond to a XFingerprint() object

        **Keyword arguments:**

        names: *List[str]*
            Names of the fingerprint as used in the database, if name == None: name = fp_type for any name in the list. 
            This parameter is used to distinguish between fingerprints in the database, i.e. it must be unique for each unique fingerprint.
            
            default: None

        force_calculate: *bool*
            Force calculation of the fingerprint even if fingerprint data was stored in the database before.

            default: False

        fingerprint_kwargs_list: *List[dict]*
            List of additional keyword arguments that are passed to each Fingerprint().__init__.

            default: None

        fingerprint_calculate_kwargs_list: *List[dict]*
            List of additional keyword arguments that are passed to each Fingerprint().calculate().

            default: None

        show_progress: *bool*
            Show a progress bar during calculation.

            default: True

        **Raises:**

        AssertionError: Number of kwargs passed to the __init__() and calculate() functions of all fingerprints are inconsitent. Thus, it is ambiguous which parameters correspond to which fingerprint.
        """
        assert len(names) <= len(fp_types), "Recieved more names than fingerprint types. Please resolve ambiguity."
        while len(fp_types) < len(names):
            names.append(None)
        for idx, fp_type in enumerate(fp_types):
            names[idx] = Fingerprint._name_from_type(names[idx], fp_type)
        if fingerprint_kwargs_list is None:
            fingerprint_kwargs_list = [{} for _ in fp_types]
        if fingerprint_calculate_kwargs_list is None:
            fingerprint_calculate_kwargs_list = [{} for _ in fp_types]
        assert len(fp_types) == len(fingerprint_kwargs_list) == len(fingerprint_calculate_kwargs_list), "Inconsistent number of kwargs! Please provide arguments for each fingerprint type or None."
        # Check if fingerprints exist
        _drop_fingerprint = []
        for idx, name in enumerate(names):
            if name in self.get_metadata().get('fingerprints', []):
                if force_calculate:
                    self.log.info(f'Fingerprints with name "{name}" exist. Overwriting...')
                else:
                    self.log.info(f'Fingerprints with name "{name}" exist. To overwrite, set `force_calculate=True`.')
                    _drop_fingerprint.append(idx)
        if len(_drop_fingerprint) > 0:
            for idx in sorted(_drop_fingerprint, reverse=True):
                fp_types.pop(idx)
                names.pop(idx)
                fingerprint_kwargs_list.pop(idx)
                fingerprint_calculate_kwargs_list.pop(idx)
        if len(fp_types) == 0:
            self.log.info("No new fingerprints to calculate.")
            return
        mid_list = []
        dictionary_list = []
        self.log.info(f"Generating {names} fingerprints...")
        for material in tqdm(self, disable = not show_progress):
            fps_data = {}
            for fp_type, name, fp_kwargs, calc_kwargs in zip(fp_types, names, fingerprint_kwargs_list, fingerprint_calculate_kwargs_list):
                new_fp = self._get_fingerprint(material, fp_type, name, fp_kwargs, force_calculate, **calc_kwargs)
                if new_fp is None:
                    continue
                else:
                    if isinstance(fp_type, type):
                        fps_data[name] = new_fp.serialize()
                    else:                               
                        fps_data[name] = new_fp.get_data_json() if isinstance(new_fp, Fingerprint) else None
            mid_list.append(material.mid)
            dictionary_list.append(fps_data)
        self.log.info("Writing fingerprints to database...")
        self.update_entries(mid_list, dictionary_list)
        self.log.info('Finished for added fingerprints.')
        self._update_metadata({'fingerprints' : names}) #TODO Update metadata schema

    def add_material(self, *args, **kwargs) -> None:
        """
        Add a specified material to the database.
        Keyword arguments are passed to the api.
        Arguments passed to this function are used to construct the database mid.
        """
        mid = self.api._gen_mid(*args)
        if self.backend.has_entry(mid):
            self.log.info(f"Material already in db: {mid}")
        else:
            material = self.api.get_calculation(*args, **kwargs)
            self.backend.add_single(material)

    def fill_database(self, *args, repeat_query: bool = False, **kwargs):
        """
        Fills the database with all materials matching the query. Parameters depend on the API that is used. 
        To perform the query even though it has been performed before, set

        ``repeat_query = True``

        See below for the documentation of the API functions.
        """
        metadata = self.get_metadata()
        try:
            query_hash = self.api.hash_query(args)
        except TypeError:
            self.log.error(f"Query can not be hashed and therefore not logged. To avoid this, please provide a `query_hash_function` to your `API`\n Traceback: {get_tb_string()}")
            query_hash = None
        if 'search_queries' in metadata.keys() and not repeat_query:
            if query_hash in metadata['search_queries']:
                self.log.info("Query has already been performed.")
                return
        self.log.info('Retrieving data...')
        materials = list(set(self.api.get_calculations_by_search(*args, **kwargs)))
        self.log.info(f"Got data for {len(materials)} entries.")
        pop_indices = []
        for idx, material in enumerate(materials):
            if self.backend.has_entry(material.mid):
                pop_indices.append(idx)
        for idx in sorted(pop_indices, reverse=True):
            mat = materials.pop(idx)
            self.log.info(f"Material {mat.mid} already in database. Skipping.")
        self.backend.add_many(materials)
        if query_hash is not None:
            self._update_metadata({'search_queries':[query_hash]})

    def get_random(self, 
                   return_mid: bool = True) -> Material:
        """
        Returns a random material from the database.
        
        **Keyword arguments:**

        return_id: *bool*
            Return id of material instead of material

        **Returns:**

        Material id (*str*) of a random entry of the database: if return_id == True
        
        Material object of a random entry of the database: else
        """
        random_id = random.randint(1, len(self))
        material = self.backend.get_by_id(random_id)
        if return_mid:
            return material.mid
        else:
            return material

    def update_entry(self, 
                     mid: str, **kwargs) -> None:
        """
        Update a single entry of the database from a given dictionary.

        Usage:

            MaterialsDatabase().update_entry("a", key = value) --> updates parameter *key* of material (with id *mid*) with *value*
        
        **Arguments:**

        mid: *str*
            Mid of the corresponding database entry

        Additional keyword arguments are used to update the database entries.
        """
        self.backend.update_single(mid, **kwargs)


    def update_entries(self, 
                       mid_list: List[str], 
                       dictionary_list: List[dict]) -> None:
        """
        Update a list of entries in the database.

        Usage:

            MaterialsDatabase().update_entries(["a", "b"], [{"key1" : value1}, {"key2" : value2}]) 
                --> updates parameter *key1* of material "a" with *value1* and *key2* of material "b" with *value2* 
        
        **Arguments:**

        mid: *str*
            Mid of the corresponding database entry

        dictionary_list: *List[dict]*
            List of dictionaries htat contains data to update
        """
        self.backend.update_many(mid_list, dictionary_list)

    def add_property(self, mid, property_name, **kwargs):
        """
        Add a property to a spacific material of the database through the API.
        
        **Arguments:**

        mid: *str*
            Id of the material to update
            
        property_name: *str*
            Name of property for storage in the database

        Keyword arguments are passed to the API to retrieve the property.
        """
        property = self.api.get_property(**self.api.resolve_mid(mid), **kwargs)
        if property is not None:
            self.update_entry(mid, **{"data" : {property_name : property}})

    def get_metadata(self) -> dict:
        """
        Get the metadata of the database.
        """
        return self.backend.metadata

    def _property_from_material(self, material, property_name):
        try:
            prop = material.get_property_by_path(property_name)
        except KeyError:
            try:
                prop = material.get_data_by_path(property_name)
            except KeyError as e:
                self.log.error(f"No property of name: {str(e)}")
                prop = None
        return prop

    def _get_fingerprint(self, material, fp_type, name, fingerprint_kwargs, force_calculate, **kwargs):
        name = Fingerprint._name_from_type(name, fp_type)
        try:
            if name in material.properties.keys() and not force_calculate:
                if isinstance(fp_type, type):
                    fp = fp_type.deserialize(material.properties[name])
                else:
                    fp = Fingerprint(fp_type, name=name).from_data(json.loads(material.properties[name]))
                fp.set_mid(material)
            else:
                fp = Fingerprint(fp_type, name = name, **fingerprint_kwargs).from_material(material, **kwargs)
            fp.set_name(name)
        except Exception as e:
            message = f"Could not get {name} fingerprint of type {fp_type} because of {str(e)}.\nTraceback: {get_tb_string()}"
            self.log.error(message)
            fp = None
        return fp

    def _update_metadata(self, update_dict):
        metadata = self.backend.metadata
        for key in update_dict.keys():
            if key in metadata.keys():
                if isinstance(update_dict[key], list):
                    if not isinstance(metadata[key], list):
                        metadata[key] = [metadata[key]]
                    for item in update_dict[key]:
                        if item not in metadata[key]:
                            metadata[key].append(item)
                else: 
                    metadata[key] = update_dict[key]
            else:
                metadata.update({key:update_dict[key]})
        self.backend.update_metadata(**metadata)

    def _init_loggers(self, path, db_filename, log_mode):

        log_mode = log_mode.strip().lower()

        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        log = logging.getLogger(db_filename +'_log')
        log.setLevel(logging.DEBUG)

        api = logging.getLogger(db_filename + '_api')
        api.setLevel(logging.DEBUG)

        handlers = []

        if log_mode != "stream":
            log_folder_name = '.' + db_filename + '_logs'
            self._log_folder_name = log_folder_name

            if not os.path.exists(os.path.join(path, log_folder_name)):
                os.makedirs(os.path.join(path, log_folder_name))

            log_file = logging.FileHandler(os.path.join(path, log_folder_name,db_filename+'_errors.log'))
            log_file.setLevel(logging.INFO)
            log_file.setFormatter(formatter)
            log_file.set_name(db_filename + '_error')
            handlers.append(log_file)
        
        if log_mode != "silent":
            console = logging.StreamHandler()
            console.setLevel(logging.DEBUG)
            console.setFormatter(formatter)
            console.set_name(db_filename + '_console')
            handlers.append(console)

        for logger in [log, api]:
            for handler in handlers:
                if handler.get_name() not in [h.get_name() for h in logger.handlers]:
                    logger.addHandler(handler)
        self.log = log
        self.api_logger = api

    def __len__(self):
        return self.backend.get_length()

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        if self._iter_index >= len(self):
            self._iter_index = 0
            raise StopIteration
        else:
            material = self.backend.get_by_id(self._iter_index)
            self._iter_index += 1
            return material

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.backend.get_single(key)
        elif isinstance(key, int):
            return self.backend.get_by_id(key)
        else:
            raise KeyError(f'Key {key} can not be interpreted as database key.')

    def __repr__(self) -> str:
        return f"MaterialsDatabase(name = {self.name}, filename = {self.backend.filename}, len = {len(self)})"


from ase import Atoms
import numpy as np

class Material():
    """
    A materials base class. Contains the atomic positions as an ASE Atoms object and additional properties as a dictionary.

    **Arguments:**

    mid: *str*
        A (unique) identifier that is used as the main identifier in a database.
    
    **Keyword arguments:**

    atoms: *ase.Atoms()*
        Atoms object describing the material

        default: *None*

    data: *dict*
        Additional data of the material
        
        default: *None*

    properties: *dict*
        Additional properies of the material

        default: *None*
    """

    def __init__(self, mid, atoms: Atoms | None = None, data = None, properties = None):
        self.mid = str(mid)
        self.set_atoms(atoms)
        self.set_data(data)
        self.set_properties(properties)

    @property
    def properties(self):
        return self._properties

    @property
    def data(self):
        return self._data

    def set_atoms(self, atoms):
        self.atoms = atoms

    def set_data(self, data):
        if data is None:
            data = {}
        self._data = data

    def set_properties(self, properties):
        if properties is None:
            properties = {}
        self._properties = properties

    def get_property_by_path(self, property_path: str):
        property_path_ = property_path.split("/")
        property_root = property_path_.pop(0)
        if property_root in self.properties.keys():
            prop = self.properties[property_root]
        else:
            raise KeyError(f"No property {property_path} for material {self.mid}")
        while len(property_path_) > 0:
            key = property_path_.pop(0)
            try:
                prop = prop[key]
            except KeyError:
                raise KeyError(f"No property {property_path} for material {self.mid}")
        return prop

    def get_data_by_path(self, data_path: str):
        data_path_ = data_path.split("/")
        data_root = data_path_.pop(0)
        if data_root in self.data.keys():
            prop = self.data[data_root]
        else:
            raise KeyError(f"No data {data_path} for material {self.mid}")
        while len(data_path_) > 0:
            key = data_path_.pop(0)
            try:
                prop = prop[key]
            except KeyError:
                raise KeyError(f"No data {data_path} for material {self.mid}")
            except TypeError as t_error:
                try:
                    prop = prop[int(key)]
                except Exception:                
                    print(f"Can't resolve: {key} of type {type(key)}. Remaining path: {data_path_}.")
                    raise t_error
        return prop

    def to_dict(self) -> dict:
        atoms_dict = self.atoms.todict()
        for key in atoms_dict.keys():
            if isinstance(atoms_dict[key], np.ndarray):
                atoms_dict[key]=atoms_dict[key].tolist()
        return {
            "mid" : self.mid,
            "atoms" : atoms_dict,
            "properties" : self.properties,
            "data" : self.data
        }

    @classmethod
    def from_dict(cls, data: dict) -> object:
        self = cls(data["mid"], 
                   atoms = Atoms.fromdict(data["atoms"]), 
                   properties = data["properties"],
                   data = data["data"])
        return self

    def __eq__(self, other: object) -> bool:
        if not self.mid == other.mid:
            return False
        if not self.atoms == other.atoms:
            return False
        if not self.data == other.data:
            return False
        if not self.properties == other.properties:
            return False
        return True

    def __repr__(self) -> str:
        return f"Material(mid = {self.mid}, data = {set(self.data.keys())}, properties = {set(self.properties.keys())})"

    def __hash__(self) -> int:
        return hash(f"{self.mid}")